import sqlite3
import os
import pandas as pd
from pm4py.objects.log.util import dataframe_utils


def apply(file_path, return_obj_dataframe=False, parameters=None):
    if parameters is None:
        parameters = {}

    db = sqlite3.connect(file_path)

    df = pd.read_sql_query("SELECT * FROM event_log", db)

    col = list(df.columns)
    rep_dict = {}
    for x in col:
        if x.startswith("case_"):
            rep_dict[x] = x.split("case_")[1]
        else:
            rep_dict[x] = "event_" + x
    df = df.rename(columns=rep_dict)

    df = df.dropna(subset=["event_id"])
    df = df.dropna(subset=["event_activity"])
    df = df.dropna(subset=["event_timestamp"])

    df = dataframe_utils.convert_timestamp_columns_in_df(df)

    """print(df)

    ot_columns = [x for x in df.columns if not x.startswith("event_")]
    for ot in ot_columns:
        df[ot] = df[ot].apply(lambda x: eval(x))

    print(df)"""

    if return_obj_dataframe:
        ot_df = pd.read_sql_query("SELECT * FROM object_types", db)
        ot_df = ot_df.dropna(subset=["NAME"])

        OT = list(ot_df["NAME"])

        obj_df_list = []

        for ot in OT:
            o_df = pd.read_sql_query("SELECT * FROM " + ot, db)

            col = list(o_df.columns)
            rep_dict = {}
            for x in col:
                rep_dict[x] = "object_"+x
            o_df = o_df.rename(columns=rep_dict)

            o_df = o_df.dropna(subset=["object_id"])
            o_df["object_type"] = ot

            o_df = dataframe_utils.convert_timestamp_columns_in_df(o_df)

            obj_df_list.append(o_df)

        obj_df = pd.concat(obj_df_list)

        return df, obj_df

    return df




from pm4pymdl.algo.mvp.utils import succint_mdl_to_exploded_mdl

def apply(df, min_acti_freq=0):
    try:
        if df.type == "succint":
            df = succint_mdl_to_exploded_mdl.apply(df)
    except:
        pass
    activ = dict(df.groupby("event_id").first()["event_activity"].value_counts())
    activ = [x for x,y in activ.items() if y >= min_acti_freq]
    return df[df["event_activity"].isin(activ)]



import networkx
from pm4py.objects.log.log import EventLog, Trace, Event
from datetime import datetime
from pm4py.objects.log.util import sorting
from collections import Counter


def apply(df, source_attr, target_attr, type_attr, timestamp_key, reverse=False):
    df[source_attr] = df[source_attr].astype(str)
    df[target_attr] = df[target_attr].astype(str)
    df[type_attr] = df[type_attr].astype(str)

    first_df = df.groupby(source_attr).first().reset_index()
    timestamps = first_df[[timestamp_key, source_attr]].to_dict('r')
    timestamps = {x[source_attr]: x[timestamp_key] for x in timestamps}

    map_types = df[[source_attr, type_attr]].to_dict('r')
    map_types = {x[source_attr]: x[type_attr] for x in map_types if x[type_attr] != 'None'}

    map_source_target = dict(df.groupby([source_attr, target_attr]).size())
    map_source_target = list(map_source_target.keys())

    map_source_target = [(map_types[x] + "=" + x, map_types[y] + "=" + y) for (x, y) in map_source_target if
                         x in map_types and y in map_types]
    map_types = {y + "=" + x: y for x, y in map_types.items()}

    G = networkx.DiGraph()
    for k in map_types:
        G.add_node(k)
    for el in map_source_target:
        if reverse:
            G.add_edge(el[1], el[0])
        else:
            G.add_edge(el[0], el[1])

    conn_comp = sorted(list(networkx.connected_components(networkx.Graph(G))), key=lambda x: len(x), reverse=True)

    return G, conn_comp, timestamps


def describe_graph(G, comp):
    SG = G.subgraph(comp)
    nodes = list(SG.nodes)
    edges = list(SG.edges)
    nodes = [(x, y) for x, y in Counter(n.split("=")[0] for n in nodes).items()]
    for i in range(len(nodes)):
        if nodes[i][1] > 1:
            nodes[i] = (nodes[i][0], "N")
        else:
            nodes[i] = (nodes[i][0], "1")
    source0 = {n[0]: [e for e in edges if e[0].split("=")[0] == n[0]] for n in nodes}
    target0 = {n[0]: [e for e in edges if e[1].split("=")[0] == n[0]] for n in nodes}
    source1 = {x: {} for x in source0}
    target1 = {x: {} for x in target0}
    for x in source0:
        ik = set(y[0] for y in source0[x])
        for k in ik:
            source1[x][k.split("=")[1]] = Counter(y[1].split("=")[0] for y in source0[x] if y[0] == k)
    for x in target0:
        ik = set(y[1] for y in target0[x])
        for k in ik:
            target1[x][k.split("=")[1]] = Counter(y[0].split("=")[0] for y in source0[x] if y[1] == k)
    print(source1)
    print(target1)
    # print(edges)


def create_log(G, conn_comp, timestamps, max_comp_len=50, include_loops=False):
    log = EventLog()
    for i in range(len(conn_comp)):
        if len(conn_comp[i]) <= max_comp_len:
            trace = Trace()
            trace.attributes["concept:name"] = str(i)
            SG = G.subgraph(conn_comp[i])
            SGG = networkx.DiGraph(SG)
            edges = list(SGG.edges)
            for e in edges:
                if e[0] == e[1]:
                    SGG.remove_edge(e[0], e[1])
            sorted_nodes = list(networkx.topological_sort(SGG))
            for n in sorted_nodes:
                selfloop = 1 if (n, n) in SG.edges else 0
                trace.append(Event({'time:timestamp': timestamps[n.split("=")[1]], 'concept:name': n.split("=")[0],
                                    'value': n.split("=")[1], 'typevalue': n, 'selfloop': selfloop}))
                if include_loops and selfloop:
                    trace.append(Event({'time:timestamp': timestamps[n.split("=")[1]], 'concept:name': n.split("=")[0],
                                        'value': n.split("=")[1], 'typevalue': n, 'selfloop': selfloop}))
            log.append(trace)
    log = sorting.sort_timestamp_log(log, "time:timestamp")
    return log



from pm4py.algo.discovery.dfg.adapters.pandas import df_statistics
from pm4py.algo.filtering.pandas.paths import paths_filter
from pm4pymdl.algo.mvp.utils import succint_mdl_to_exploded_mdl
from pm4py.util.constants import PARAMETER_CONSTANT_CASEID_KEY, PARAMETER_CONSTANT_ATTRIBUTE_KEY
import pandas as pd


def filter_paths(df, paths, parameters=None):
    """
    Apply a filter on traces containing / not containing a path

    Parameters
    ----------
    df
        Dataframe
    paths
        Paths to filter on
    parameters
        Possible parameters of the algorithm, including:
            case_id_glue -> Case ID column in the dataframe
            attribute_key -> Attribute we want to filter
            positive -> Specifies if the filter should be applied including traces (positive=True)
            or excluding traces (positive=False)
    Returns
    ----------
    df
        Filtered dataframe
    """
    try:
        if df.type == "succint":
            df = succint_mdl_to_exploded_mdl.apply(df)
    except:
        pass
    if parameters is None:
        parameters = {}
    paths = [path[0] + "," + path[1] for path in paths]
    case_id_glue = parameters[
        PARAMETER_CONSTANT_CASEID_KEY] if PARAMETER_CONSTANT_CASEID_KEY in parameters else CASE_CONCEPT_NAME
    attribute_key = parameters[
        PARAMETER_CONSTANT_ATTRIBUTE_KEY] if PARAMETER_CONSTANT_ATTRIBUTE_KEY in parameters else DEFAULT_NAME_KEY
    df = df.sort_values([case_id_glue, "event_timestamp"])
    positive = parameters["positive"] if "positive" in parameters else True
    filt_df = df[[case_id_glue, attribute_key, "event_id"]]
    filt_dif_shifted = filt_df.shift(-1)
    filt_dif_shifted.columns = [str(col) + '_2' for col in filt_dif_shifted.columns]
    stacked_df = pd.concat([filt_df, filt_dif_shifted], axis=1)
    stacked_df["@@path"] = stacked_df[attribute_key] + "," + stacked_df[attribute_key + "_2"]
    stacked_df = stacked_df[stacked_df["@@path"].isin(paths)]
    i1 = df.set_index("event_id").index
    i2 = stacked_df.set_index("event_id").index
    i3 = stacked_df.set_index("event_id_2").index
    if positive:
        return df[i1.isin(i2) | i1.isin(i3)]
    else:
        return df[~i1.isin(i2) & ~i1.isin(i3)]


def apply(df, min_freq=0):
    if min_freq > 0:
        persps = [x for x in df.columns if not x.startswith("event_")]
        collation = []
        for persp in persps:
            red_df = df.dropna(subset=[persp])
            prevlen = len(df)
            while True:
                dfg = df_statistics.get_dfg_graph(red_df, activity_key="event_activity", timestamp_key="event_timestamp", case_id_glue=persp)
                dfg = [x for x in dfg if dfg[x] >= min_freq]
                param = {}
                param[PARAMETER_CONSTANT_CASEID_KEY] = persp
                param[PARAMETER_CONSTANT_ATTRIBUTE_KEY] = "event_activity"
                red_df = filter_paths(red_df, dfg, parameters=param)
                thislen = len(red_df)
                dfg = df_statistics.get_dfg_graph(red_df, activity_key="event_activity", timestamp_key="event_timestamp", case_id_glue=persp)
                if len(dfg) == 0 or min(dfg.values()) >= min_freq or prevlen == thislen:
                    collation.append(red_df)
                    break
                prevlen = thislen
        return pd.concat(collation)
    return df



from pm4pymdl.algo.mvp.utils import filter_metaclass
from pm4pymdl.algo.mvp.utils import succint_mdl_to_exploded_mdl


def filter_float(df, act, attr, v1, v2, parameters=None):
    if parameters is None:
        parameters = {}

    try:
        if df.type == "succint":
            df = succint_mdl_to_exploded_mdl.apply(df)
    except:
        pass

    red_df = df[df["event_activity"] == act]
    red_df = red_df[df[attr] >= v1]
    red_df = red_df[df[attr] <= v2]

    red_df = red_df.dropna(how="all", axis=1)

    return filter_metaclass.do_filtering(df, red_df, parameters=parameters)




import pandas as pd


def perfom_cleaning(df, parameters=None):
    if parameters is None:
        parameters = {}
    allowed_activities = parameters["allowed_activities"] if "allowed_activities" in parameters else None
    if allowed_activities is not None:
        persps = [x for x in df.columns if not x.startswith("event_")]
        collation = []
        for persp in persps:
            red_df = df.dropna(subset=[persp])
            if persp in allowed_activities:
                red_df = red_df[red_df["event_activity"].isin(allowed_activities[persp])]
                collation.append(red_df)
        if collation:
            df = pd.concat(collation)
        else:
            df = pd.DataFrame()
    return df



from pm4pymdl.algo.mvp.utils import exploded_mdl_to_grouped_stream_old

def apply(df, dest_path, parameters=None):
    """
    Export into a XOC format

    Parameters
    ------------
    df
        Dataframe
    dest_path
        Destination path
    parameters
        Parameters of the algorithm
    """
    if parameters is None:
        parameters = {}

    grouped_stream = exploded_mdl_to_grouped_stream_old.apply(df)

    F = open(dest_path, "w")
    F.write("<?xml version=\"1.0\" encoding=\"UTF-8\" ?>\n")
    F.write("<log xes.features=\"nested-attributes\">\n")
    for eve_id in grouped_stream:
        first = grouped_stream[eve_id][0]
        F.write("\t<event>\n")
        F.write("\t\t<string key=\"id\" value=\"" + str(eve_id) + "\" />\n")
        F.write("\t\t<string key=\"activity\" value=\"" + first["event_activity"] + "\" />\n")
        F.write("\t\t<string key=\"timestamp\" value=\"" + first["event_timestamp"].strftime(
            "%Y-%m-%d %H:%M:%S") + "\" />\n")
        F.write("\t\t<references>\n")
        for obj in grouped_stream[eve_id]:
            obj_id = None
            obj_class = None
            for prop in obj.keys():
                if not prop.startswith("event_"):
                    obj_class = prop
                    obj_id = obj[prop]
            F.write("\t\t\t<object>\n")
            F.write("\t\t\t\t<string key=\"id\" value=\""+str(obj_id)+"\" />\n")
            F.write("\t\t\t\t<string key=\"class\" value=\""+obj_class+"\" />\n")
            F.write("\t\t\t</object>\n")
        F.write("\t\t</references>\n")
        F.write("\t\t<model>\n")
        F.write("\t\t\t<objects>\n")
        for obj in grouped_stream[eve_id]:
            obj_id = None
            obj_class = None
            for prop in obj.keys():
                if not prop.startswith("event_"):
                    obj_class = prop
                    obj_id = obj[prop]
            F.write("\t\t\t\t<object>\n")
            F.write("\t\t\t\t\t<string key=\"id\" value=\""+str(obj_id)+"\" />\n")
            F.write("\t\t\t\t\t<string key=\"class\" value=\""+obj_class+"\" />\n")
            F.write("\t\t\t\t</object>\n")
        F.write("\t\t\t</objects>\n")
        F.write("\t\t</model>\n")
        F.write("\t</event>\n")
        pass
    F.write("</log>\n")
    F.close()



import pandas as pd

def apply(file_path, return_obj_dataframe=False, parameters=None):
    if parameters is None:
        parameters = {}

    sep = parameters["sep"] if "sep" in parameters else ","
    quotechar = parameters["quotechar"] if "quotechar" in parameters else "\""

    if file_path.endswith(".csv") or file_path.endswith(".mdl"):
        all_df = pd.read_csv(file_path, sep=sep, quotechar=quotechar)
        eve_cols = [x for x in all_df.columns if not x.startswith("object_")]
        obj_cols = [x for x in all_df.columns if x.startswith("object_")]
        df = all_df[eve_cols]
        obj_df = pd.DataFrame()
        if obj_cols:
            obj_df = all_df[obj_cols]
        df["event_timestamp"] = pd.to_datetime(df["event_timestamp"])
        if "event_start_timestamp" in df.columns:
            df["event_start_timestamp"] = pd.to_datetime(df["event_start_timestamp"])
        df = df.dropna(subset=["event_id"])
        df["event_id"] = df["event_id"].astype(str)
        df.type = "succint"
    elif file_path.endswith(".parquet"):
        from pm4pymdl.util.parquet_importer import importer as parquet_importer
        all_df = parquet_importer.apply(file_path)
        eve_cols = [x for x in all_df.columns if not x.startswith("object_")]
        obj_cols = [x for x in all_df.columns if x.startswith("object_")]
        df = all_df[eve_cols]
        df.type = "exploded"
        obj_df = pd.DataFrame()
        if obj_cols:
            obj_df = all_df[obj_cols]
        df["event_timestamp"] = pd.to_datetime(df["event_timestamp"])
        df = df.dropna(subset=["event_id"])
        df.type = "exploded"

    if return_obj_dataframe:
        obj_df = obj_df.dropna(subset=["object_id"])
        return df, obj_df

    return df



def apply(df, path, parameters=None):
    """
    Exports a dataframe to a Parquet file

    Parameters
    ------------
    df
        Dataframe
    path
        Path
    parameters
        Possible parameters of the algorithm
    """
    if parameters is None:
        parameters = {}

    df.to_parquet(path)



from pm4pymdl.algo.mvp.utils import succint_mdl_to_exploded_mdl


def apply(df, persp, parameters=None):
    if parameters is None:
        parameters = {}

    try:
        if df.type == "succint":
            df = succint_mdl_to_exploded_mdl(df)
            df.type = "exploded"
    except:
        pass

    cols = [x for x in df.columns if x.startswith("event_") or x == persp]
    df = df[cols].dropna(subset=[persp])

    return dict(df.groupby("event_id").first()["event_activity"].value_counts())



from pm4pymdl.algo.mvp.utils import succint_mdl_to_exploded_mdl


def apply(df, persp, parameters=None):
    if parameters is None:
        parameters = {}

    try:
        if df.type == "succint":
            df = succint_mdl_to_exploded_mdl(df)
            df.type = "exploded"
    except:
        pass

    cols = [x for x in df.columns if x.startswith("event_") or x == persp]
    df = df[cols].dropna(subset=[persp])

    df = df.groupby(["event_activity", persp]).first()
    values = df.groupby(["event_id", "event_activity"]).size().to_dict()

    ret = {}
    for v in values:
        if not v[1] in ret:
            ret[v[1]] = []
        ret[v[1]].append(values[v])

    for k in ret:
        ret[k] = sorted(ret[k])

    return ret



import pandas as pd
from datetime import datetime
import pytz


def get_dt_from_string(dt):
    """
    If the date is expressed as string, do the conversion to a datetime.datetime object

    Parameters
    -----------
    dt
        Date (string or datetime.datetime)

    Returns
    -----------
    dt
        Datetime object
    """
    if type(dt) is str:
        return datetime.strptime(dt, "%Y-%m-%d %H:%M:%S")

    return dt


def apply(dataframe, dt1, dt2, parameters=None):
    if parameters is None:
        parameters = {}

    dt1 = get_dt_from_string(dt1)
    dt2 = get_dt_from_string(dt2)
    #dt1 = dt1.replace(tzinfo=pytz.utc)
    #dt2 = dt2.replace(tzinfo=pytz.utc)
    #dt1 = pd.to_datetime(dt1, utc=True)
    #dt2 = pd.to_datetime(dt2, utc=True)

    dataframe = dataframe[dataframe["event_timestamp"] > dt1]
    dataframe = dataframe[dataframe["event_timestamp"] < dt2]

    return dataframe



from pm4pymdl.algo.mvp.gen_framework.edge_freq import type11, type12, type13, type211, type212, type221, type222, type231, type232

TYPE11 = "type11"
TYPE12 = "type12"
TYPE13 = "type13"
TYPE211 = "type211"
TYPE212 = "type212"
TYPE221 = "type221"
TYPE222 = "type222"
TYPE231 = "type231"
TYPE232 = "type232"

VERSIONS = {TYPE11: type11.apply, TYPE12: type12.apply, TYPE13: type13.apply, TYPE211: type211.apply,
            TYPE212: type212.apply, TYPE221: type221.apply, TYPE222: type222.apply, TYPE231: type231.apply,
            TYPE232: type232.apply}


def apply(df, model, rel_ev, rel_act, variant=TYPE11, parameters=None):
    return VERSIONS[variant](df, model, rel_ev, rel_act, parameters=parameters)



from collections import Counter


def get(dataframe, activity, ot, parameters=None):
    if parameters is None:
        parameters = {}

    red_df = dataframe[dataframe["event_activity"] == activity]
    red_df = red_df.dropna(axis=1, how="all")
    red_df = red_df.dropna(subset=[ot])

    lista = [y for x,y in dict(red_df.groupby("event_id").size()).items()]

    return dict(Counter(lista))




import uuid
import random
from datetime import datetime
import numpy as np
from copy import copy


def generate_stream(n_obj=10000, n_eve=100000, n_classes=3, n_activities=40, avg_obj_per_class=1.0):
    activities = []
    for i in range(n_activities):
        activities.append(str(uuid.uuid4()))
    classes_objects = {}
    for i in range(n_classes):
        classes_objects[str(uuid.uuid4())] = list()
    classes = list(classes_objects.keys())
    for i in range(n_obj):
        cl = random.choice(classes)
        classes_objects[cl].append(str(uuid.uuid4()))
    timestamp = 10000000
    discovery_stream = []
    conformance_stream = []
    for i in range(n_eve):
        basic_eve = {"event_id": str(i), "event_activity": random.choice(activities), "event_timestamp": datetime.fromtimestamp(timestamp)}
        eve = copy(basic_eve)
        for cl in classes:
            num_obj = max(min(int(round(avg_obj_per_class - 1 + np.random.exponential())), int(round(len(classes_objects[cl])/2))), 1)
            objs = list(random.sample(classes_objects[cl], num_obj))
            eve[cl] = objs
            for o in objs:
                red_eve = copy(basic_eve)
                red_eve[cl] = o
                discovery_stream.append(red_eve)
        conformance_stream.append(eve)
        timestamp = timestamp + 1
    return discovery_stream, conformance_stream




from pm4pymdl.algo.mvp.gen_framework4.versions_discovery.classic import get_stream_from_dataframe, apply_stream


def apply(df, model, parameters=None):
    if parameters is None:
        parameters = {}

    stream = get_stream_from_dataframe(df, parameters=parameters)
    return apply_stream(stream, model, parameters=parameters)


def apply_stream(stream, model, parameters=None):
    log_model = apply_stream(stream, parameters=parameters)



def get(dataframe, parameters=None):
    if parameters is None:
        parameters = {}

    r0 = dataframe["event_timestamp"].describe()

    ret = {x: str(y) for x, y in r0.items()}

    return ret



from pm4pymdl.algo.mvp.utils import exploded_mdl_to_succint_mdl
from collections import Counter
import math
from pm4py.objects.dfg.filtering import dfg_filtering


class Shared:
    diff_values = []


def count_objs(values):
    if type(values) is str and values[0] == "[":
        values = eval(values)
    if type(values) is list:
        return len(values)
    return 0


def apply_bound(value):
    for i in range(len(Shared.diff_values)):
        if value in Shared.diff_values[i]:
            return str(min(Shared.diff_values[i])) + "<=x<=" + str(max(Shared.diff_values[i]))
    # raise Exception("sub error", value, Shared.diff_values)
    return None


def convert_to_stream(df):
    stream = df.to_dict('r')
    for i in range(len(stream)):
        stream_keys = sorted(list(stream[i].keys()))
        for j in range(len(stream_keys)):
            vv = str(stream[i][stream_keys[j]]).lower()
            if vv == "nan" or vv == "nat" or vv == "none" or vv.startswith("0"):
                del stream[i][stream_keys[j]]
    return stream


def get_bindings_from_log(succint_table, max_no_divisions=2, noise_threshold=0.2, parameters=None):
    if parameters is None:
        parameters = {}

    try:
        if succint_table.type == "exploded":
            succint_table = exploded_mdl_to_succint_mdl.apply(succint_table)
    except:
        pass

    obj_count = succint_table.copy()
    OT = [x for x in obj_count.columns if not x.startswith("event_")]

    values_ot = {}
    values_act = {}

    for ot in OT:
        obj_count[ot] = obj_count[ot].apply(count_objs)
        all_values = sorted(list(set(obj_count[ot].dropna().unique())))
        if 0 in all_values:
            del all_values[all_values.index(0)]
        divv = math.ceil(len(all_values) / max_no_divisions)
        diff_values = [[all_values[j] for j in range(len(all_values)) if j // divv == i] for i in
                       range(max_no_divisions)]
        diff_values = [x for x in diff_values if len(x) > 0]
        Shared.diff_values = diff_values
        values_ot[ot] = diff_values
        obj_count[ot] = obj_count[ot].apply(apply_bound)

    stream = convert_to_stream(obj_count)
    activities = set(x["event_activity"] for x in stream)

    acti_dict = {
        x: [{a: b for a, b in y.items() if not a.startswith("event_")} for y in stream if y["event_activity"] == x] for
        x in activities}

    for act in activities:
        values_act[act] = {}

        acti_count = Counter()
        initial_dfg = Counter()

        dic = acti_dict[act]
        for item in dic:
            for key in item:
                acti_count[str((key, item[key]))] += 1
                for other_key in item:
                    if key != other_key:
                        initial_dfg[(str((key, item[key])), str((other_key, item[other_key])))] += 1

        acti_count = dict(acti_count)
        initial_dfg = dict(initial_dfg)

        values_act[act]["acti_count"] = acti_count
        values_act[act]["initial_dfg"] = initial_dfg

        dfg = dfg_filtering.clean_dfg_based_on_noise_thresh(initial_dfg, acti_count, noise_threshold)
        diff_dfg = {x: y for x, y in initial_dfg.items() if x not in dfg}

        values_act[act]["dfg"] = dfg
        values_act[act]["diff_dfg"] = diff_dfg

    ret = {}
    ret["ot"] = values_ot
    ret["act"] = values_act

    return ret

    #print(acti_dict)



from pm4pymdl.algo.mvp.utils import succint_mdl_to_exploded_mdl
import pandas as pd


def do_filtering(dataframe, fd0, parameters=None):
    if parameters is None:
        parameters = {}

    try:
        if dataframe.type == "succint":
            dataframe = succint_mdl_to_exploded_mdl.apply(dataframe)
    except:
        pass

    fd1 = dataframe[dataframe["event_id"].isin(fd0["event_id"])]
    cols = [x for x in fd1.columns if not x.startswith("event_")]
    cols_values = {}
    collation = []
    for c in cols:
        cols_values[c] = list(fd1.dropna(subset=[c])[c].unique())
        collation.append(dataframe[dataframe[c].isin(cols_values[c])])

    if collation:
        df2 = pd.concat(collation)
    else:
        df2 = pd.DataFrame()

    return df2


def do_negative_filtering(dataframe, fd0, parameters=None):
    if parameters is None:
        parameters = {}

    try:
        if dataframe.type == "succint":
            dataframe = succint_mdl_to_exploded_mdl.apply(dataframe)
    except:
        pass

    fd1 = dataframe[dataframe["event_id"].isin(fd0["event_id"])]
    cols = [x for x in fd1.columns if not x.startswith("event_")]
    cols_values = {}
    collation = []
    for c in cols:
        cols_values[c] = list(fd1.dropna(subset=[c])[c].unique())
        collation.append(dataframe[dataframe[c].isin(cols_values[c])])

    if collation:
        df2 = pd.concat(collation)
    else:
        df2 = pd.DataFrame()

    i1 = dataframe.index
    i2 = df2.index

    return dataframe[~i1.isin(i2)]



from pm4pymdl.algo.mvp.utils import succint_mdl_to_exploded_mdl, clean_objtypes
import pandas as pd

def preprocess(df, parameters=None):
    if parameters is None:
        parameters = {}

    conversion_needed = False

    try:
        if df.type == "succint":
            conversion_needed = True
    except:
        pass

    if len(df) == 0:
        df = pd.DataFrame({"event_id": [], "event_activity": []})

    if conversion_needed:
        df = succint_mdl_to_exploded_mdl.apply(df)

    #df = clean_objtypes.perfom_cleaning(df, parameters=parameters)

    if len(df) == 0:
        df = pd.DataFrame({"event_id": [], "event_activity": []})

    return df



import os
import shutil
import subprocess
import sys
import traceback


from pm4pymdl.visualization.mvp.versions import classic

CLASSIC = "classic"

VERSIONS = {CLASSIC: classic.apply}


def apply(model, parameters=None, variant=CLASSIC):
    """
    Visualize StarStar models

    Parameters
    ------------
    model
        StarStar model
    parameters
        Possible parameters of the algorithm
    variant
        Variants available for the visualizer, possible values: classic

    Returns
    -----------
    file_name
        File name where a representation of the StarStar model is stored
    """
    return VERSIONS[variant](model, parameters=parameters)


def view(figure):
    """
    View on the screen a figure that has been rendered

    Parameters
    ----------
    figure
        figure
    """
    try:
        filename = figure.name
        figure = filename
    except AttributeError:
        # continue without problems, a proper path has been provided
        pass

    is_ipynb = False

    try:
        get_ipython()
        is_ipynb = True
    except NameError:
        pass

    if is_ipynb:
        from IPython.display import Image
        return Image(open(figure, "rb").read())
    else:
        if sys.platform.startswith('darwin'):
            subprocess.call(('open', figure))
        elif os.name == 'nt':  # For Windows
            os.startfile(figure)
        elif os.name == 'posix':  # For Linux, Mac, etc.
            subprocess.call(('xdg-open', figure))


def save(figure, output_file_path):
    """
    Save a figure that has been rendered

    Parameters
    -----------
    figure
        figure
    output_file_path
        Path where the figure should be saved
    """
    try:
        filename = figure.name
        figure = filename
    except AttributeError:
        # continue without problems, a proper path has been provided
        pass
        traceback.print_exc()

    shutil.copyfile(figure, output_file_path)



import json
import pandas as pd
from dateutil import parser


def apply(file_path, return_obj_df=None, parameters=None):
    if parameters is None:
        parameters = {}

    prefix = "jmd:"
    F = open(file_path, "rb")
    obj = json.load(F)
    F.close()
    eve_stream = obj[prefix+"events"]
    for el in eve_stream:
        eve_stream[el]["event_id"] = el
    obj_stream = obj[prefix+"objects"]
    for el in obj_stream:
        obj_stream[el]["object_id"] = el
    obj_stream = list(obj_stream.values())
    obj_type = {}
    for el in obj_stream:
        obj_type[el["object_id"]] = el[prefix+"otyp"]
        del el[prefix+"otyp"]
        for k2 in el[prefix+"ovmap"]:
            el["object_"+k2] = el[prefix+"ovmap"][k2]
        del el[prefix+"ovmap"]
    eve_stream = list(eve_stream.values())
    for el in eve_stream:
        new_omap = {}
        for obj in el[prefix+"omap"]:
            typ = obj_type[obj]
            if not typ in new_omap:
                new_omap[typ] = set()
            new_omap[typ].add(obj)
        for typ in new_omap:
            new_omap[typ] = list(new_omap[typ])
        el[prefix+"omap"] = new_omap
        el["event_activity"] = el[prefix+"activity"]
        el["event_timestamp"] = parser.parse(el[prefix+"timestamp"])
        del el[prefix+"activity"]
        del el[prefix+"timestamp"]
        for k2 in el[prefix+"vmap"]:
            el["event_"+k2] = el[prefix+"vmap"][k2]
        del el[prefix+"vmap"]
        for k2 in el[prefix+"omap"]:
            el[k2] = el[prefix+"omap"][k2]
        del el[prefix + "omap"]

    eve_df = pd.DataFrame(eve_stream)
    obj_df = pd.DataFrame(obj_stream)

    eve_df.type = "succint"

    if return_obj_df or (return_obj_df is None and len(obj_df.columns) > 1):
        return eve_df, obj_df
    return eve_df





from pm4pymdl.algo.mvp.utils import exploded_mdl_to_stream


def apply(df, model, parameters=None):
    if parameters is None:
        parameters = {}

    stream = get_conf_stream_from_dataframe(df, parameters=parameters)

    return apply_stream(stream, model, parameters=parameters)


def get_conf_stream_from_dataframe(df, parameters=None):
    if parameters is None:
        parameters = {}

    df_type = df.type
    df = df.sort_values(["event_timestamp", "event_id"])
    if df_type == "exploded":
        stream = exploded_mdl_to_stream.apply(df)
    else:
        stream = df.to_dict('r')

    for ev in stream:
        keys = list(ev.keys())
        for k in keys:
            if str(ev[k]) == "nan":
                del ev[k]
            elif not k.startswith("event_") and type(ev[k]) == str and ev[k][0] == "[":
                ev[k] = eval(ev[k])

    return stream


def apply_stream(stream, model, parameters=None):
    if parameters is None:
        parameters = {}

    ret = []
    model = model.dictio

    must_start = {}
    must_end = {}
    must_edges = {}
    dictio_objs = {}
    min_obj = {}
    max_obj = {}

    for t in model["types_view"]:
        dictio_objs[t] = {}
        must_edges[t] = {}
        min_obj[t] = {}
        max_obj[t] = {}
        for ak in model["types_view"][t]["start_activities"]:
            if model["types_view"][t]["start_activities"][ak]["must"]:
                if t not in must_start:
                    must_start[t] = set()
                must_start[t].add(ak)
        for ak in model["types_view"][t]["end_activities"]:
            if model["types_view"][t]["end_activities"][ak]["must"]:
                if t not in must_end:
                    must_end[t] = set()
                must_end[t].add(ak)
        for ek in model["types_view"][t]["edges"]:
            e = model["types_view"][t]["edges"][ek]
            if e["must"]:
                must_edges[t][ek[1]] = ek[0]
            min_obj[t][ek] = e["min_obj"]
            max_obj[t][ek] = e["max_obj"]

    for i in range(len(stream)):
        ret.append(set())
        ev = stream[i]
        class_keys = [k for k in ev.keys() if not k.startswith("event_")]
        for t in class_keys:
            for o in ev[t]:
                if not o in dictio_objs[t]:
                    if t in must_start and not ev["event_activity"] in must_start[t]:
                        ret[-1].add("start activity for object %s of type %s - should be %s instead is %s" % (
                            o, t, must_start[t], ev["event_activity"]))
                else:
                    prev = dictio_objs[t][o]
                    this_edge = (prev["event_activity"], ev["event_activity"])
                    rel_ev_obj = set(o2 for o2 in ev[t] if o2 in dictio_objs[t] and dictio_objs[t][o2] == prev)
                    if not (min_obj[t][this_edge] <= len(rel_ev_obj) <= max_obj[t][this_edge]):
                        ret[-1].add(
                            "number of related objects of type %s should be between %d and %d - instead is %d" % (
                                t, min_obj[t][this_edge], max_obj[t][this_edge], len(rel_ev_obj)))

                    if t in must_end and prev["event_activity"] in must_end[t]:
                        ret[-1].add(
                            "object %s of type %s - %s should be an end activity, instead is followed by %s!" % (
                                o, t, prev["event_activity"], ev["event_activity"]))
                    else:
                        if this_edge[1] in must_edges[t]:
                            if not must_edges[t][this_edge[1]] == this_edge[0]:
                                ret[-1].add(
                                    "object %s of type %s - %s should be preceded by %s, instead is preceded by %s!" % (
                                        o, t, this_edge[1], must_edges[t][this_edge[1]], this_edge[0]))
            for o in ev[t]:
                dictio_objs[t][o] = ev
    return ret



import bson
import json
import pandas as pd
import os


def convert(source_path, target_path):
    from pm4pymdl.util.parquet_exporter import exporter as parquet_exporter
    limit_length = 35
    codec_options = bson.CodecOptions(unicode_decode_error_handler='ignore')
    gen = bson.decode_file_iter(open(source_path, 'rb'), codec_options=codec_options)
    json_list = []
    df_list = []
    i = 0
    for row in gen:
        cols = list(row.keys())
        for col in cols:
            if type(row[col]) is dict:
                if "id" in row[col]:
                    row[col] = row[col]["id"]
                else:
                    del row[col]
            elif type(row[col]) is str and len(row[col]) > limit_length:
                del row[col]
        json_list.append(row)
        if i > 0 and i % 10000 == 0:
            print(i)
            df = pd.DataFrame(json_list)
            df_list.append(df)
            json_list = None
            json_list = []
        i = i + 1
    if json_list:
        df = pd.DataFrame(json_list)
        df_list.append(df)
    if df_list:
        overall_df = pd.concat(df_list)
        for col in overall_df.columns:
            overall_df[col] = overall_df[col].astype(str)
        parquet_exporter.apply(overall_df, target_path, parameters={"compression": "gzip"})


folder_content = os.listdir("msr14")
for file in folder_content:
    if file.endswith(".bson") and not file == "commits.bson":
        print("converting ",file)
        source_path = os.path.join("msr14", file)
        target_path = os.path.join("output", file)
        convert(source_path, target_path)



import random
import time
from copy import copy

from pm4py.algo.discovery.dfg.adapters.pandas import df_statistics
from pm4py.algo.filtering.pandas.attributes import attributes_filter
from pm4py.algo.filtering.pandas.end_activities import end_activities_filter
from pm4py.algo.filtering.pandas.start_activities import start_activities_filter
from pm4py.objects.heuristics_net.net import HeuristicsNet
from pm4py.util import constants
import pandas as pd
from collections import Counter

DEPENDENCY_THRESH = "dependency_thresh"
AND_MEASURE_THRESH = "and_measure_thresh"
MIN_ACT_COUNT = "min_act_count"
MIN_DFG_OCCURRENCES = "min_dfg_occurrences"
DFG_PRE_CLEANING_NOISE_THRESH = "dfg_pre_cleaning_noise_thresh"
LOOP_LENGTH_TWO_THRESH = "loop_length_two_thresh"

DEFAULT_DEPENDENCY_THRESH = -0.9
DEFAULT_AND_MEASURE_THRESH = 0.65
DEFAULT_MIN_ACT_COUNT = 1
DEFAULT_MIN_DFG_OCCURRENCES = 1
DEFAULT_DFG_PRE_CLEANING_NOISE_THRESH = -0.05
DEFAULT_LOOP_LENGTH_TWO_THRESH = 0.5

DECREASING_FACTOR = "decreasingFactor"
PERFORMANCE = "performance"
PERSPECTIVES = "perspectives"
USE_TIMESTAMP = "use_timestamp"
SORT_CASEID_REQUIRED = "sort_caseid_required"
SORT_TIMESTAMP_REQUIRED = "sort_timestamp_required"

COLORS = ["#05B202", "#A13CCD", "#39F6C0", "#BA0D39", "#E90638", "#07B423", "#306A8A", "#678225", "#2742FE", "#4C9A75",
          "#4C36E9", "#7DB022", "#EDAC54", "#EAC439", "#EAC439", "#1A9C45", "#8A51C4", "#496A63", "#FB9543", "#2B49DD",
          "#13ADA5", "#2DD8C1", "#2E53D7", "#EF9B77", "#06924F", "#AC2C4D", "#82193F", "#0140D3"]


def clean_sa_ea(dictio, decreasing_factor):
    """
    Clean start and end activities by using decreasing factor

    Parameters
    -------------
    dictio
        Dictionary of start and end activities
    decreasing_factor
        Decreasing factor

    Returns
    -------------
    cleaned_dictio
        Cleaned dictionary
    """
    cleaned_dictio = {}
    ordered_list = sorted([(x, y) for x, y in dictio.items()], key=lambda x: x[1], reverse=True)
    i = 0
    while i < len(ordered_list):
        if i == 0:
            cleaned_dictio[ordered_list[i][0]] = ordered_list[i][1]
        else:
            ratio = ordered_list[i][1] / ordered_list[i - 1][1]
            if ratio >= decreasing_factor:
                cleaned_dictio[ordered_list[i][0]] = ordered_list[i][1]
            else:
                break
        i = i + 1
    return cleaned_dictio


def apply(df, parameters=None):
    """
    Discover a StarStar model from an ad-hoc built dataframe

    Parameters
    -------------
    df
        Dataframe
    parameters
        Possible parameters of the algorithm

    Returns
    -------------
    perspectives_heu
        Dictionary of perspectives associated to Heuristics Net
    """

    if parameters is None:
        parameters = {}

    if len(df) == 0:
        df = pd.DataFrame({"event_id": [], "event_activity": []})

    dependency_thresh = parameters[
        DEPENDENCY_THRESH] if DEPENDENCY_THRESH in parameters else DEFAULT_DEPENDENCY_THRESH
    and_measure_thresh = parameters[
        AND_MEASURE_THRESH] if AND_MEASURE_THRESH in parameters else DEFAULT_AND_MEASURE_THRESH
    min_act_count = parameters[MIN_ACT_COUNT] if MIN_ACT_COUNT in parameters else DEFAULT_MIN_ACT_COUNT
    min_dfg_occurrences = parameters[
        MIN_DFG_OCCURRENCES] if MIN_DFG_OCCURRENCES in parameters else DEFAULT_MIN_DFG_OCCURRENCES
    dfg_pre_cleaning_noise_thresh = parameters[
        DFG_PRE_CLEANING_NOISE_THRESH] if DFG_PRE_CLEANING_NOISE_THRESH in parameters else DEFAULT_DFG_PRE_CLEANING_NOISE_THRESH
    decreasing_factor_sa_ea = parameters[DECREASING_FACTOR] if DECREASING_FACTOR in parameters else 0.5
    performance = parameters[PERFORMANCE] if PERFORMANCE in parameters else False
    perspectives = parameters[PERSPECTIVES] if PERSPECTIVES in parameters else None
    use_timestamp = parameters[USE_TIMESTAMP] if USE_TIMESTAMP in parameters else True
    sort_caseid_required = parameters[SORT_CASEID_REQUIRED] if SORT_CASEID_REQUIRED in parameters else True
    sort_timestamp_required = parameters[SORT_TIMESTAMP_REQUIRED] if SORT_TIMESTAMP_REQUIRED in parameters else True

    perspectives_heu = {}

    r = lambda: random.randint(0, 255)
    if perspectives is None:
        perspectives = list(x for x in df.columns if not x.startswith("event"))

        # del perspectives[perspectives.index("event_id")]
        # del perspectives[perspectives.index("event_activity")]
        # if "event_timestamp" in perspectives:
        #    del perspectives[perspectives.index("event_timestamp")]
        perspectives = sorted(perspectives)
    activities_occurrences = attributes_filter.get_attribute_values(df.groupby("event_id").first().reset_index(),
                                                                    "event_activity")
    activities_occurrences = {x: y for x, y in activities_occurrences.items() if y >= min_act_count}
    for p_ind, p in enumerate(perspectives):
        has_timestamp = False
        if "event_timestamp" in df.columns and use_timestamp:
            proj_df = df[["event_id", "event_activity", "event_timestamp", p]].dropna(subset=[p])
            has_timestamp = True
        else:
            proj_df = df[["event_id", "event_activity", p]].dropna(subset=[p])

        """
        proj_df_activities = Counter(list(proj_df["event_activity"]))
        proj_df_activities = set(x for x in proj_df_activities if proj_df_activities[x] >= min_act_count)
        proj_df = proj_df[proj_df["event_activity"].isin(proj_df_activities)]
        print(proj_df_activities)"""

        # proj_df = proj_df.groupby(["event_id", "event_activity", p]).first().reset_index()
        # print('sii')

        proj_df = proj_df[proj_df["event_activity"].isin(activities_occurrences)]
        proj_df_first = proj_df.groupby("event_id").first().reset_index()

        if performance:
            dfg_frequency, dfg_performance = df_statistics.get_dfg_graph(proj_df, activity_key="event_activity",
                                                                         case_id_glue=p,
                                                                         timestamp_key="event_timestamp",
                                                                         measure="both",
                                                                         sort_caseid_required=sort_caseid_required,
                                                                         sort_timestamp_along_case_id=sort_timestamp_required)
        else:
            if has_timestamp:
                dfg_frequency = df_statistics.get_dfg_graph(proj_df, activity_key="event_activity", case_id_glue=p,
                                                            timestamp_key="event_timestamp",
                                                            sort_caseid_required=sort_caseid_required,
                                                            sort_timestamp_along_case_id=sort_timestamp_required)
            else:
                dfg_frequency = df_statistics.get_dfg_graph(proj_df, activity_key="event_activity", case_id_glue=p,
                                                            sort_timestamp_along_case_id=False,
                                                            sort_caseid_required=sort_caseid_required)
        if len(dfg_frequency) > 0:
            this_color = COLORS[p_ind] if p_ind < len(COLORS) else '#%02X%02X%02X' % (r(), r(), r())
            parameters_sa_ea = copy(parameters)
            parameters_sa_ea[constants.PARAMETER_CONSTANT_CASEID_KEY] = p
            parameters_sa_ea[constants.PARAMETER_CONSTANT_ATTRIBUTE_KEY] = "event_activity"
            parameters_sa_ea[constants.PARAMETER_CONSTANT_ACTIVITY_KEY] = "event_activity"
            start_activities = start_activities_filter.get_start_activities(proj_df, parameters=parameters_sa_ea)
            end_activities = end_activities_filter.get_end_activities(proj_df, parameters=parameters_sa_ea)
            start_activities = clean_sa_ea(start_activities, decreasing_factor_sa_ea)
            end_activities = clean_sa_ea(end_activities, decreasing_factor_sa_ea)
            max_entry = dict()
            max_exit = dict()
            for x in dfg_frequency:
                a1 = x[0]
                a2 = x[1]
                y = dfg_frequency[x]
                if not a1 == a2:
                    if not a2 in max_entry or max_entry[a2][1] < y:
                        max_entry[a2] = [x, y]
                    if not a1 in max_exit or max_exit[a1][1] < y:
                        max_exit[a1] = [x, y]
            max_entry = set(y[0] for y in max_entry.values())
            max_exit = set(y[0] for y in max_exit.values())
            #max_entry = {}
            #max_exit = {}
            dfg_frequency = {x: y for x, y in dfg_frequency.items() if
                             y >= min_dfg_occurrences or x in max_entry or x in max_exit}
            activities = list(activities_occurrences.keys())

            if performance:
                dfg_performance = {x: y for x, y in dfg_performance.items() if x in dfg_frequency}
                heu_net = HeuristicsNet(dfg_frequency, start_activities=start_activities, end_activities=end_activities,
                                        default_edges_color=this_color, net_name=p, activities=activities,
                                        activities_occurrences=activities_occurrences, performance_dfg=dfg_performance)
            else:
                heu_net = HeuristicsNet(dfg_frequency, start_activities=start_activities, end_activities=end_activities,
                                        default_edges_color=this_color, net_name=p, activities=activities,
                                        activities_occurrences=activities_occurrences)
            heu_net.calculate(dependency_thresh=dependency_thresh, and_measure_thresh=and_measure_thresh,
                              min_act_count=1, min_dfg_occurrences=1,
                              dfg_pre_cleaning_noise_thresh=dfg_pre_cleaning_noise_thresh)
            if len(heu_net.nodes) > 0:
                perspectives_heu[p] = heu_net

    return perspectives_heu



import pyarrow.parquet as pq
from pm4py.objects.log.util import dataframe_utils
import deprecation
from pm4py.util import exec_utils
from pm4pymdl.util.parquet_importer.parameters import Parameters

COLUMNS = Parameters.COLUMNS.value


def apply(path, parameters=None):
    """
    Import a Parquet file

    Parameters
    -------------
    path
        Path of the file to import
    parameters
        Parameters of the algorithm, possible values:
            Parameters.COLUMNS -> columns to import from the Parquet file

    Returns
    -------------
    df
        Pandas dataframe
    """
    if parameters is None:
        parameters = {}

    columns = exec_utils.get_param_value(Parameters.COLUMNS, parameters, None)

    if columns:
        df = pq.read_pandas(path, columns=columns).to_pandas()
    else:
        df = pq.read_pandas(path).to_pandas()

    return dataframe_utils.legacy_parquet_support(df)


def import_table(path, parameters=None):
    """
    Imports a Pyarrow table from a Parquet file

    Parameters
    ------------
    path
        Path to the Parquet file
    parameters
        Parameters of the algorithm

    Returns
    ------------
    table
        Pyarrow table
    """
    if parameters is None:
        parameters = {}

    columns = exec_utils.get_param_value(Parameters.COLUMNS, parameters, None)

    if columns:
        table = pq.read_table(path, columns=columns)
    else:
        table = pq.read_table(path)

    return table


def import_log(path, parameters=None):
    """
    Imports an event log from a Parquet file

    Parameters
    --------------
    path
        Path to the Parquet file
    parameters
        Parameters of the algorithm

    Returns
    --------------
    log
        Event log
    """
    if parameters is None:
        parameters = {}

    table = import_table(path, parameters=parameters)

    return dataframe_utils.table_to_log(table, parameters=parameters)



from pm4py.algo.discovery.dfg.utils.dfg_utils import get_max_activity_count, get_activities_from_dfg
from pm4py.algo.filtering.common import filtering_constants
from pm4py.util import constants

def clean_dfg_based_on_noise_thresh(dfg, activities, noise_threshold, parameters=None):
    """
    Clean Directly-Follows graph based on noise threshold

    Parameters
    ----------
    dfg
        Directly-Follows graph
    activities
        Activities in the DFG graph
    noise_threshold
        Noise threshold

    Returns
    ----------
    newDfg
        Cleaned dfg based on noise threshold
    """
    if parameters is None:
        parameters = {}

    most_common_paths = parameters[
        constants.PARAM_MOST_COMMON_PATHS] if constants.PARAM_MOST_COMMON_PATHS in parameters else None
    if most_common_paths is None:
        most_common_paths = []

    new_dfg = None

    for el in dfg:
        if new_dfg is None:
            new_dfg = {}
        act1 = el.split("@@")[0]
        act2 = el.split("@@")[1]
        val = dfg[el]

        if not el in most_common_paths and val < min(activities[act1] * noise_threshold,
                                                     activities[act2] * noise_threshold):
            pass
        else:
            new_dfg[el] = dfg[el]

    if new_dfg is None:
        return dfg

    return new_dfg



import tempfile
import uuid
from graphviz import Digraph
from pm4py.objects.process_tree import pt_operator
from pm4pymdl.visualization.mvp.gen_framework2.versions import util

COLORS = ["#05B202", "#A13CCD", "#39F6C0", "#BA0D39", "#E90638", "#07B423", "#306A8A", "#678225", "#2742FE", "#4C9A75",
          "#4C36E9", "#7DB022", "#EDAC54", "#EAC439", "#EAC439", "#1A9C45", "#8A51C4", "#496A63", "#FB9543", "#2B49DD",
          "#13ADA5", "#2DD8C1", "#2E53D7", "#EF9B77", "#06924F", "#AC2C4D", "#82193F", "#0140D3"]


def repr_tree(tree, viz, current_node, rec_depth, res, acti_map, persp_color, edges_map, activ_freq_map, measure, freq,
              key, classifier, node_shape, projection, parameters):
    """
    Represent a subtree on the GraphViz object

    Parameters
    -----------
    tree
        Current subtree
    viz
        GraphViz object
    current_node
        Father node of the current subtree
    rec_depth
        Reached recursion depth
    parameters
        Possible parameters of the algorithm.

    Returns
    -----------
    gviz
        (partial) GraphViz object
    """
    freq_prefix = "E="
    if freq == "objects":
        freq_prefix = "O="
    elif freq == "eo":
        freq_prefix = "EO="

    for child in tree.children:
        if child.operator is None:
            this_trans_id = str(uuid.uuid4())
            if child.label is None:
                viz.node(this_trans_id, "tau", style='filled', color=persp_color, fontcolor=persp_color,
                         fillcolor="white", shape='box', fixedsize='true', width="2.5",
                     fontsize="8")
            else:
                if not str(child) in acti_map:
                    acti_map[str(child)] = this_trans_id
                    if str(child) in res["activities_repeated"]:
                        viz.node(this_trans_id, str(child) + " (" + freq_prefix + str(res["activities"][str(child)]) + ")",
                                 style="filled", fillcolor="white", shape=node_shape, fixedsize='true', width="2.5",
                     fontsize="8")
                    else:
                        viz.node(this_trans_id, str(child) + " (" + freq_prefix + str(res["activities"][str(child)]) + ")",
                                 style="filled", fillcolor=persp_color, shape=node_shape, fixedsize='true', width="2.5",
                     fontsize="8")
                else:
                    this_trans_id = acti_map[str(child)]
            viz.edge(current_node, this_trans_id, color=persp_color)
        else:
            condition_wo_operator = child.operator == pt_operator.Operator.XOR and len(
                child.children) == 1 and child.children[0].operator is None
            if condition_wo_operator:
                childchild = child.children[0]
                viz.attr('node', shape='box', fixedsize='true', width="2.5",
                         fontsize="8")
                this_trans_id = str(uuid.uuid4())
                if childchild.label is None:
                    viz.node(this_trans_id, str(childchild), style='filled', color=persp_color, fontcolor=persp_color,
                             fillcolor="white", shape='box', fixedsize='true', width="2.5",
                     fontsize="8")
                else:
                    if not str(childchild) in acti_map:
                        acti_map[str(childchild)] = this_trans_id
                        if str(childchild) in res["activities_repeated"]:
                            viz.node(this_trans_id, str(childchild) + " (" + freq_prefix + str(
                                activ_freq_map[key][str(childchild)]) + ")", style="filled", fillcolor="white", shape=node_shape, fixedsize='true', width="2.5",
                     fontsize="8")
                        else:
                            viz.node(this_trans_id, str(childchild) + " (" + freq_prefix + str(
                                activ_freq_map[key][str(childchild)]) + ")", style="filled", fillcolor=persp_color, shape=node_shape, fixedsize='true', width="2.5",
                     fontsize="8")
                    else:
                        this_trans_id = acti_map[str(childchild)]
                viz.edge(current_node, this_trans_id, color=persp_color)
            else:
                op_node_identifier = str(uuid.uuid4())
                viz.node(op_node_identifier, str(child.operator), shape="circle", fixedsize='true', width="0.6",
                         fontsize="14", style="filled", fillcolor=persp_color)
                viz.edge(current_node, op_node_identifier, color=persp_color)
                viz, acti_map = repr_tree(child, viz, op_node_identifier, rec_depth + 1, res, acti_map, persp_color,
                                          edges_map, activ_freq_map,
                                          measure, freq, key, classifier, node_shape, projection, parameters)
    return viz, acti_map


def apply(res, measure="frequency", freq="events", classifier="activity", projection="no", parameters=None):
    if parameters is None:
        parameters = {}

    min_edge_freq = parameters["min_edge_freq"] if "min_edge_freq" in parameters else 0

    filename = tempfile.NamedTemporaryFile(suffix='.gv')
    viz = Digraph("pt", filename=filename.name, engine='dot', graph_attr={'bgcolor': 'transparent'})
    image_format = parameters["format"] if "format" in parameters else "png"
    acti_map = {}

    freq_prefix = "E="
    if freq == "objects":
        freq_prefix = "O="
    elif freq == "eo":
        freq_prefix = "EO="

    node_shape = "box" if classifier == "activity" else "trapezium"

    reference_map = {}
    events_map = {}

    edges_map = {}
    activ_freq_map = {}

    for key in res["models"]:
        edges_map[key] = util.get_edges_map(key, res, measure=measure, variant=freq)
        activ_freq_map[key] = util.get_activity_map_frequency(key, res, variant=freq)
        reference_map[key] = util.get_activity_map_frequency(key, res, variant="objects")
        events_map[key] = util.get_edges_map(key, res, measure="frequency", variant="events")

    edges_map = util.projection_edges_freq(edges_map, events_map, min_edge_freq)
    edges_map = util.projection(edges_map, reference_map, type=projection)

    count = 0
    for key, tree in res["models"].items():
        persp_color = COLORS[count % len(COLORS)]
        count = count + 1

        # add first operator
        if tree.operator:
            op_node_identifier = str(uuid.uuid4())
            viz.node(op_node_identifier, str(tree.operator), shape='circle', fixedsize='true', width="0.6",
                     fontsize="14", style="filled", fillcolor=persp_color)

            viz, acti_map = repr_tree(tree, viz, op_node_identifier, 0, res, acti_map, persp_color, edges_map,
                                      activ_freq_map, measure, freq, key, classifier, node_shape, projection, parameters)
        else:
            this_trans_id = str(uuid.uuid4())
            if tree.label is None:
                viz.node(this_trans_id, "tau", style='filled', color=persp_color, fontcolor=persp_color,
                         fillcolor="white", shape='box', fixedsize='true', width="2.5",
                     fontsize="8")
            else:
                if not str(tree) in acti_map:
                    acti_map[str(tree)] = this_trans_id
                    if str(tree) in res["activities_repeated"]:
                        viz.node(this_trans_id,
                                 str(tree) + " (" + freq_prefix + str(activ_freq_map[key][str(tree)]) + ")",
                                 style="filled", fillcolor="white", shape=node_shape, fixedsize='true', width="2.5",
                     fontsize="8")
                    else:
                        viz.node(this_trans_id,
                                 str(tree) + " (" + freq_prefix + str(activ_freq_map[key][str(tree)]) + ")",
                                 style="filled", fillcolor=persp_color, shape=node_shape, fixedsize='true', width="2.5",
                     fontsize="8")
                else:
                    this_trans_id = acti_map[str(tree)]

    viz.attr(overlap='false')
    viz.attr(fontsize='11')
    viz.format = image_format

    return viz



import tempfile
import uuid
from graphviz import Digraph
from pm4py.objects.process_tree import pt_operator
from pm4pymdl.visualization.mvp.gen_framework2.versions import util


COLORS = ["#05B202", "#A13CCD", "#39F6C0", "#BA0D39", "#E90638", "#07B423", "#306A8A", "#678225", "#2742FE", "#4C9A75",
          "#4C36E9", "#7DB022", "#EDAC54", "#EAC439", "#EAC439", "#1A9C45", "#8A51C4", "#496A63", "#FB9543", "#2B49DD",
          "#13ADA5", "#2DD8C1", "#2E53D7", "#EF9B77", "#06924F", "#AC2C4D", "#82193F", "#0140D3"]


def apply(res, measure="frequency", freq="events", classifier="activity", projection="no", parameters=None):
    if parameters is None:
        parameters = {}

    min_edge_freq = parameters["min_edge_freq"] if "min_edge_freq" in parameters else 0

    filename = tempfile.NamedTemporaryFile(suffix='.gv')
    viz = Digraph("pt", filename=filename.name, engine='dot', graph_attr={'bgcolor': 'transparent'})
    image_format = parameters["format"] if "format" in parameters else "png"
    acti_map = {}

    freq_prefix = "E="
    if freq == "objects":
        freq_prefix = "O="
    elif freq == "eo":
        freq_prefix = "EO="

    node_shape = "box" if classifier == "activity" else "trapezium"

    reference_map = {}
    events_map = {}

    edges_map = {}
    activ_freq_map = {}

    for key in res["models"]:
        edges_map[key] = util.get_edges_map(key, res, measure=measure, variant=freq)
        activ_freq_map[key] = util.get_activity_map_frequency(key, res, variant=freq)
        reference_map[key] = util.get_activity_map_frequency(key, res, variant="objects")
        events_map[key] = util.get_edges_map(key, res, measure="frequency", variant="events")

    edges_map = util.projection_edges_freq(edges_map, events_map, min_edge_freq)
    edges_map = util.projection(edges_map, reference_map, type=projection)

    count = 0

    for key, model in res["models"].items():
        persp_color = COLORS[count % len(COLORS)]
        count = count + 1

        net, im, fm = model

        # places
        viz.attr('node', shape='circle', fixedsize='true', width='0.75')
        for p in net.places:
            viz.node(str(id(p)), "", style="filled", fillcolor=persp_color)

        for t in net.transitions:
            trans_id = str(id(t))
            if t.label is None:
                viz.node(trans_id, "", style='filled', fillcolor="black", color=persp_color, shape='box', width='3.8')
            elif t.label in acti_map:
                trans_id = acti_map[t.label]
            else:
                if t.label in res["activities_repeated"]:
                    viz.node(trans_id, t.label+" ("+freq_prefix+str(activ_freq_map[key][t.label])+")", style="filled", fillcolor="white", shape=node_shape, width='3.8')
                else:
                    viz.node(trans_id, t.label+" ("+freq_prefix+str(activ_freq_map[key][t.label])+")", style="filled", fillcolor=persp_color, shape=node_shape, width='3.8')
                acti_map[t.label] = trans_id
            for arc in t.in_arcs:
                p = arc.source
                viz.edge(str(id(p)), trans_id, color=persp_color)
            for arc in t.out_arcs:
                p = arc.target
                viz.edge(trans_id, str(id(p)), color=persp_color)

    viz.attr(overlap='false')
    viz.attr(fontsize='11')
    viz.format = image_format

    return viz



from pm4pymdl.algo.mvp.utils import exploded_mdl_to_succint_mdl, succint_mdl_to_exploded_mdl
import pandas as pd
import json


def json_serial(obj):
    """JSON serializer for objects not serializable by default json code"""
    if "time" in str(type(obj)):
        stru = str(obj)
        stru = stru.replace(" ","T") + "Z"
        return stru
    return str(obj)


def get_type(t0):
    if "float" in str(t0).lower() or "double" in str(t0).lower():
        return "double"
    elif "object" in str(t0).lower():
        return "string"
    else:
        return "string"


def apply(df, file_path, obj_df=None, parameters=None):
    if parameters is None:
        parameters = {}

    prefix = "jmd:"

    conversion_needed = True
    try:
        if df.type == "succint":
            conversion_needed = False
    except:
        pass

    if conversion_needed:
        df = exploded_mdl_to_succint_mdl.apply(df)

    if obj_df is not None:
        df = pd.concat([df, obj_df])

    col = list(df.columns)
    rep_dict = {}
    obj_types = set()

    for x in col:
        if x.startswith("event_"):
            rep_dict[x] = x.split("event_")[1]
        else:
            y = "case_" + x
            rep_dict[x] = y
            obj_types.add(x)
    df = df.rename(columns=rep_dict)
    df = df.dropna(subset=["id"])
    df = df.dropna(subset=["activity"])
    df = df.dropna(subset=["timestamp"])
    df["id"] = df["id"].astype(str)

    activities = list(df["activity"].unique())

    acti_df = {}
    ot_df = {}

    acti_mandatory = {}
    ot_mandatory = {}
    objects_from_df_type = {}

    att_types = {}

    for act in activities:
        red_df = df[df["activity"] == act].dropna(how="all", axis=1)
        red_df2 = red_df.dropna(how="any", axis=1)
        acti_df[act] = red_df

        for col in red_df.columns:
            if not col.startswith("case_") and not col in ["id", "activity", "timestamp"]:
                att_types[col] = get_type(red_df[col].dtype)

        acti_mandatory[act] = []
        for col in red_df2.columns:
            if not col.startswith("case_") and not col in ["id", "activity", "timestamp"]:
                acti_mandatory[act].append(col)

    if obj_df is not None:
        OT = obj_df["object_type"].dropna().unique()

        for ot in OT:
            obj_types.add(ot)

            red_df = obj_df[obj_df["object_type"] == ot].dropna(how="all", axis=1)

            red_df = red_df.dropna(subset=["object_id"]).drop(columns=["object_type"])

            col = list(red_df.columns)
            rep_dict = {}
            for x in col:
                rep_dict[x] = x.split("object_")[1]
            red_df = red_df.rename(columns=rep_dict)

            for col in red_df.columns:
                if not col in ["id", "type"]:
                    att_types[col] = get_type(red_df[col].dtype)

            red_df2 = red_df.dropna(how="any", axis=1)

            ot_mandatory[ot] = []
            for col in red_df2.columns:
                if not col in ["id"]:
                    #type_mandatory[ot][col] = get_type(red_df2[col].dtype)
                    ot_mandatory[ot].append(col)

            ot_df[ot] = red_df

    ret = {}
    att_names = sorted(list(set(att_types.keys())))
    att_typ_values = sorted(list(set(att_types.values())))
    object_types = sorted(list(obj_types))
    ret[prefix+"att_names"] = att_names
    ret[prefix+"att_types"] = att_typ_values
    ret[prefix+"object_types"] = object_types
    ret[prefix+"types_corr"] = att_types
    ret[prefix+"events"] = {}
    ret[prefix+"objects"] = {}
    ret[prefix+"att_mand"] = acti_mandatory
    ret[prefix+"ot_mand"] = ot_mandatory

    stream = df.dropna(how="all", axis=1).to_dict("r")
    for el in stream:
        el2 = {}
        el2[prefix + "activity"] = el["activity"]
        el2[prefix + "timestamp"] = el["timestamp"]
        el2[prefix + "omap"] = {}
        el2[prefix + "vmap"] = {}

        for k in el:
            try:
                if el[k] is not None and str(el[k]).lower() != "nan" and str(el[k]).lower() != "nat":
                    if k.startswith("case_"):
                        if type(el[k]) is str:
                            y = eval(el[k])
                        else:
                            y = el[k]
                        if y:
                            if len(y) > 1 or len(y[0]) > 0:
                                el2[prefix + "omap"][k.split("case_")[1]] = y
                            for subel in y:
                                objects_from_df_type[subel] = k.split("case_")[1]
                    elif not k in ["id", "activity", "timestamp"]:
                        el2[prefix + "vmap"][k] = el[k]
            except:
                pass
        el2[prefix + "omap"] = list(set(z for y in el2[prefix+"omap"].values() for z in y))
        ret[prefix + "events"][el["id"]] = el2

    for t in ot_df:
        stream = ot_df[t].to_dict('r')
        for el in stream:
            el2 = {}
            el2[prefix + "otyp"] = t
            el2[prefix + "ovmap"] = {}
            for k in el:
                if not k in ["id", "type"]:
                    el2[prefix + "ovmap"][k] = el[k]
            ret[prefix + "objects"][el["id"]] = el2
    for o in objects_from_df_type:
        if not o in ret[prefix + "objects"]:
            ret[prefix + "objects"][o] = {prefix+"otyp": objects_from_df_type[o], prefix+"ovmap": {}}
    F = open(file_path, "w")
    json.dump(ret, F, default=json_serial, indent=2)
    F.close()



def calculate_lead_time(succint_dataframe):
    if len(succint_dataframe) > 0:
        ev_timest = succint_dataframe[["event_start_timestamp", "event_timestamp"]].to_dict('r')
        ev_timest = sorted([(x["event_start_timestamp"].timestamp(), x["event_timestamp"].timestamp()) for x in ev_timest])
        return max(x[1] for x in ev_timest) - ev_timest[0][0]
    return 0.0


def calculate_cycle_time(succint_dataframe):
    if len(succint_dataframe) > 0:
        cycle = 0.0
        ev_timest = succint_dataframe[["event_start_timestamp", "event_timestamp"]].to_dict('r')
        ev_timest = sorted(
            [(x["event_start_timestamp"].timestamp(), x["event_timestamp"].timestamp()) for x in ev_timest])
        curr_min = ev_timest[0][0]
        curr_max = ev_timest[0][1]
        i = 1
        while i < len(ev_timest):
            this_min = ev_timest[i][0]
            this_max = ev_timest[i][1]

            if this_min > curr_max:
                cycle += (curr_max - curr_min)
                curr_min = this_min
                curr_max = this_max
            else:
                curr_max = max(curr_max, this_max)
            i = i +1
        cycle += (curr_max - curr_min)
        return cycle
    return 0.0



from pm4pymdl.algo.mvp.gen_framework2 import md_petri_inductive, md_petri_alpha, md_process_trees, md_dfg, md_ts


def apply(df, variant="dfg", classifier_function=None, parameters=None):
    if parameters is None:
        parameters = {}

    if variant == "inductive_petri":
        return md_petri_inductive.discovery.apply(df, classifier_function=classifier_function, parameters=parameters)
    elif variant == "alpha_petri":
        return md_petri_alpha.discovery.apply(df, classifier_function=classifier_function, parameters=parameters)
    elif variant == "process_tree":
        return md_process_trees.discovery.apply(df, classifier_function=classifier_function, parameters=parameters)
    elif variant == "dfg":
        return md_dfg.discovery.apply(df, classifier_function=classifier_function, parameters=parameters)
    elif variant == "trans_system":
        return md_ts.discovery.apply(df, classifier_function=classifier_function, parameters=parameters)



from frozendict import frozendict


def create_index(succint_dataframe):
    events = succint_dataframe.to_dict("r")
    index = {}
    for idx, ev in enumerate(events):
        keys = [x for x in ev.keys() if not x.startswith("event_")]
        for k in keys:
            if not k in index:
                index[k] = {}
            events[idx][k] = tuple(eval(events[idx][k]))
        # events[idx] = frozendict(events[idx])

    for ev in events:
        keys = [x for x in ev.keys() if not x.startswith("event_")]
        for k in keys:
            for val in ev[k]:
                if not val in index[k]:
                    index[k][val] = set()
                index[k][val].add(ev["event_id"])

    events_dict = {x["event_id"]: x for x in events}

    return events_dict, index


def apply(events_dict, index, event, provided_keys=None):
    curr = [event["event_id"]]
    i = 0
    while i < len(curr):
        ev = events_dict[curr[i]]
        this_timestamp = ev["event_timestamp"]
        keys = set(x for x in ev.keys() if not x.startswith("event_"))
        if provided_keys is not None:
            keys = keys.intersection(provided_keys)
        for k in keys:
            for val in ev[k]:
                for evv in index[k][val]:
                    if not evv in curr:
                        if events_dict[evv]["event_timestamp"] > this_timestamp:
                            curr.append(evv)
        i = i + 1
    return set(curr)


def filter_dataframe(dataframe, events_dict, index, event, provided_keys=None):
    X_result = apply(events_dict, index, event, provided_keys=provided_keys)
    return dataframe[dataframe["event_id"].isin(X_result)]


def signature(events_dict, index, event, provided_keys=None):
    X_result = apply(events_dict, index, event, provided_keys=provided_keys)
    sign = dict()

    for evk in X_result:
        ev = events_dict[evk]
        keys = set(x for x in ev.keys() if not x.startswith("event_"))
        if provided_keys is not None:
            keys = keys.intersection(provided_keys)
        for k in keys:
            if not k in sign or sign[k] < len(ev[k]):
                sign[k] = len(ev[k])

    sign = tuple(sorted([(x, y) for x, y in sign.items()]))
    return sign


def equivalence_classes(events_dict, index, activity_to_consider=None, provided_keys=None):
    classes = {}
    for evk in events_dict:
        ev = events_dict[evk]
        if activity_to_consider is None or ev["event_activity"] == activity_to_consider:
            sign = signature(events_dict, index, ev, provided_keys=provided_keys)
            if not sign in classes:
                classes[sign] = set()
            classes[sign].add(evk)
    return classes



from pm4pymdl.algo.mvp.utils import exploded_mdl_to_grouped_stream_old


def apply(df, remove_common=False, include_activity_timest_in_key=False):
    """
    Dataframe to grouped stream

    Parameters
    -------------
    df
        Dataframe

    Returns
    -------------
    grouped_stream
        Grouped stream of events
    """

    grouped_stream = exploded_mdl_to_grouped_stream_old.apply(df, remove_common=remove_common,
                                                              include_activity_timest_in_key=include_activity_timest_in_key)

    ret = []

    for gk in grouped_stream:
        g = grouped_stream[gk]
        if g:
            ev = {}
            keys = set()
            for i in range(len(g)):
                keys = keys.union(set(g[i].keys()))
            for key in keys:
                if key.startswith("event_"):
                    ev[key] = [g[i][key] for i in range(len(g)) if key in g[i]][0]
                else:
                    ev[key] = [g[i][key] for i in range(len(g)) if key in g[i]]
            ret.append(ev)

    return ret



import pandas

from pm4py.objects.conversion.log.variants import to_data_frame as to_dataframe
from pm4pymdl.util.parquet_exporter.versions import pandas as pandas_exporter
from enum import Enum
from pm4py.util import exec_utils


DEFAULT_VARIANT = pandas_exporter


class Variants(Enum):
    PANDAS = pandas_exporter


try:
    from pm4pymdl.util.parquet_exporter.versions import pyarrow

    # enums cannot be changed
    class Variants(Enum):
        PANDAS = pandas_exporter
        PYARROW = pyarrow
except:
    # pyarrow is not installed correctly or does not work
    pass


def apply(log, path, parameters=None, variant=DEFAULT_VARIANT):
    """
    Exports a log to a Parquet file

    Parameters
    ------------
    log
        Log
    path
        Path
    parameters
        Possible parameters of the algorithm
    variant
        Variant of the algorithm, possible values:
            - Variants.PYARROW
            - Variants.PANDAS
    """
    if not type(log) is pandas.core.frame.DataFrame:
        log = to_dataframe.apply(log)

    return exec_utils.get_variant(variant).apply(log, path, parameters=parameters)


def export_log(log, path, parameters=None, variant=DEFAULT_VARIANT):
    """
    Exports a log to a Parquet file

    Parameters
    ------------
    log
        Log
    path
        Path
    parameters
        Possible parameters of the algorithm
    variant
        Variant of the algorithm, possible values:
            - Variants.PYARROW
            - Variants.PANDAS
    """
    if not type(log) is pandas.core.frame.DataFrame:
        log = to_dataframe.apply(log)

    return exec_utils.get_variant(variant).apply(log, path, parameters=parameters)


def export_df(log, path, parameters=None, variant=DEFAULT_VARIANT):
    """
    Exports a log to a Parquet file

    Parameters
    ------------
    log
        Log
    path
        Path
    parameters
        Possible parameters of the algorithm
    variant
        Variant of the algorithm, possible values:
            - Variants.PYARROW
            - Variants.PANDAS
    """
    if not type(log) is pandas.core.frame.DataFrame:
        log = to_dataframe.apply(log)

    return exec_utils.get_variant(variant).apply(log, path, parameters=parameters)



def get(dataframe, activity, parameters=None):
    if parameters is None:
        parameters = {}

    ret = {}

    red_df = dataframe[dataframe["event_activity"] == activity]
    red_df = red_df.groupby("event_id").first()
    red_df = red_df.dropna(axis=1, how="all")
    cols = [x for x in red_df if x.startswith("event_") and not x in ["event_id", "event_activity", "event_timestamp"]]
    for c in cols:
        ret[c] = str(red_df[c].dtype)

    return ret



from pm4pymdl.algo.mvp.gen_framework.models.basic_model_structure import BasicModelStructure


class MVPModel2(BasicModelStructure):
    def __init__(self, df, parameters=None):
        self.map = None
        self.type = "model2"
        BasicModelStructure.__init__(self, df, parameters=parameters)

    def get_map(self):
        return self.map

    def set_map(self, map):
        self.map = map

    def calculate_map(self):
        self.map = {}
        activities = self.df["event_activity"].unique()
        nnc = self.df.count()

        for act in activities:
            red_df = self.df[self.df["event_activity"] == act]

            non_null_count = dict(red_df.count())

            max_key = None
            max_value = -1

            keys = sorted([key for key in non_null_count.keys() if not key.startswith("event_")])

            for key in keys:
                if non_null_count[key] > max_value:
                    max_key = key
                    max_value = non_null_count[key]
                elif non_null_count[key] >= max_value and nnc[key] > nnc[max_key]:
                    max_key = key
                    max_value = non_null_count[key]

            self.map[act] = max_key



from pm4pymdl.algo.mvp.utils import succint_mdl_to_exploded_mdl
from pm4py.objects.conversion.log import converter
import math
from statistics import mean


def apply(df, parameters=None):
    if parameters is None:
        parameters = {}

    stream = get_stream_from_dataframe(df, parameters=parameters)

    return apply_stream(stream, parameters=parameters)


def get_stream_from_dataframe(df, parameters=None):
    if parameters is None:
        parameters = {}

    df_type = df.type
    df = df.sort_values(["event_timestamp", "event_id"])
    if df_type == "succint":
        df = succint_mdl_to_exploded_mdl.apply(df)

    columns = [x for x in df.columns if not x.startswith("event") or x == "event_activity" or x == "event_id" or x == "event_timestamp"]
    df = df[columns]

    stream = converter.apply(df, variant=converter.Variants.TO_EVENT_STREAM)

    return stream


def apply_stream(stream, parameters=None):
    if parameters is None:
        parameters = {}

    support = parameters["support"] if "support" in parameters else 1
    epsilon = parameters["epsilon"] if "epsilon" in parameters else 0.0
    debug = parameters["debug"] if "debug" in parameters else False
    noise_obj_number = parameters["noise_obj_number"] if "noise_obj_number" in parameters else 0.0

    types_lifecycle = {}

    eo = {}
    eot = {}
    eoe = {}
    ee = {}
    timestamps = {}

    for ev in stream:
        cl = [k for k in ev if not k.startswith("event_") and str(ev[k]) != "nan"][0]
        if not cl in types_lifecycle:
            types_lifecycle[cl] = {}
        o = ev[cl]
        if not o in types_lifecycle[cl]:
            types_lifecycle[cl][o] = []
        types_lifecycle[cl][o].append(ev)

    for t in types_lifecycle:
        eot[t] = dict()
        objects_lifecycle = types_lifecycle[t]
        for o in objects_lifecycle:
            evs = objects_lifecycle[o]
            i = 0
            while i < len(evs):
                i1 = evs[i]["event_id"]
                a1 = evs[i]["event_activity"]
                t1 = evs[i]["event_timestamp"].timestamp()
                if a1 not in eo:
                    eo[a1] = set()
                if a1 not in eot[t]:
                    eot[t][a1] = set()
                eo[a1].add((i1, o))
                eot[t][a1].add((i1, o))
                if i < len(evs)-1:
                    i2 = evs[i+1]["event_id"]
                    a2 = evs[i+1]["event_activity"]
                    t2 = evs[i+1]["event_timestamp"].timestamp()
                    if not (a1, t, a2) in eoe:
                        eoe[(a1, t, a2)] = set()
                        ee[(a1, t, a2)] = set()
                    eoe[(a1, t, a2)].add((i1, o, i2))
                    ee[(a1, t, a2)].add((i1, i2))
                    if not (i1, o, i2) in timestamps:
                        timestamps[(i1, o, i2)] = []
                    if not (i1, i2) in timestamps:
                        timestamps[(i1, i2)] = []
                    timestamps[(i1, o, i2)].append(t2-t1)
                    timestamps[(i1, i2)].append(t2-t1)
                i = i + 1

    for el in timestamps:
        timestamps[el] = mean(timestamps[el])

    ret = {}
    ret["activities"] = {}
    for act in eo:
        ret["activities"][act] = {}
        ret["activities"][act]["events"] = len({x[0] for x in eo[act]})
        ret["activities"][act]["objects"] = len({x[1] for x in eo[act]})
        ret["activities"][act]["eo"] = len(eo[act])
    ret["types_view"] = {}
    activities_mapping = {}
    activities_mapping_count = {}
    for t in types_lifecycle:
        ret["types_view"][t] = {"edges": {}, "activities": {}}
        for act in eot[t]:
            values = eot[t][act]
            ret["types_view"][t]["activities"][act] = {}
            ret["types_view"][t]["activities"][act]["events"] = {x[0] for x in values}
            ret["types_view"][t]["activities"][act]["objects"] = {x[1] for x in values}
            ret["types_view"][t]["activities"][act]["eo"] = values
        available_keys = {x for x in eoe.keys() if x[1] == t}
        for k in available_keys:
            a1 = k[0]
            a2 = k[2]
            values = eoe[k]
            values_ee = ee[k]
            values_timestamp_eoe = mean([timestamps[v] for v in values])
            values_timestamp_ee = mean([timestamps[v] for v in values_ee])
            g_1_2 = group_1_2(values)
            g_1_3 = group_1_3(values)
            g_1_2 = g_1_2[:math.ceil(len(g_1_2)*(1.0-noise_obj_number))]
            g_1_3 = g_1_3[:math.ceil(len(g_1_3)*(1.0-noise_obj_number))]
            g_1_2_min = min(g_1_2)
            g_1_2_max = max(g_1_2)
            g_1_3_min = min(g_1_3)
            g_1_3_max = max(g_1_3)
            ret["types_view"][t]["edges"][(a1, a2)] = {}
            ret["types_view"][t]["edges"][(a1, a2)]["events"] = {(x[0], x[2]) for x in values}
            ret["types_view"][t]["edges"][(a1, a2)]["objects"] = {x[1] for x in values}
            ret["types_view"][t]["edges"][(a1, a2)]["eo"] = values
            ret["types_view"][t]["edges"][(a1, a2)]["support_entry"] = ret["types_view"][t]["activities"][a2]["objects"].intersection(ret["types_view"][t]["edges"][(a1, a2)]["objects"])
            ret["types_view"][t]["edges"][(a1, a2)]["dev_entry"] = ret["types_view"][t]["activities"][a2]["objects"].difference(ret["types_view"][t]["edges"][(a1, a2)]["objects"])
            ret["types_view"][t]["edges"][(a1, a2)]["support_exit"] = ret["types_view"][t]["activities"][a1]["objects"].intersection(ret["types_view"][t]["edges"][(a1, a2)]["objects"])
            ret["types_view"][t]["edges"][(a1, a2)]["dev_exit"] = ret["types_view"][t]["activities"][a1]["objects"].difference(ret["types_view"][t]["edges"][(a1, a2)]["objects"])
            sen = len(ret["types_view"][t]["edges"][(a1, a2)]["support_entry"])
            den = len(ret["types_view"][t]["edges"][(a1, a2)]["dev_entry"])
            sex = len(ret["types_view"][t]["edges"][(a1, a2)]["support_exit"])
            dex = len(ret["types_view"][t]["edges"][(a1, a2)]["dev_exit"])
            if sen >= support and den/sen <= epsilon:
                ret["types_view"][t]["edges"][(a1, a2)]["must_entry"] = True
            else:
                ret["types_view"][t]["edges"][(a1, a2)]["must_entry"] = False
            if sex >= support and dex/sex <= epsilon:
                ret["types_view"][t]["edges"][(a1, a2)]["must_exit"] = True
            else:
                ret["types_view"][t]["edges"][(a1, a2)]["must_exit"] = False
            ret["types_view"][t]["edges"][(a1, a2)]["min_exit_obj"] = g_1_2_min
            ret["types_view"][t]["edges"][(a1, a2)]["max_exit_obj"] = g_1_2_max
            ret["types_view"][t]["edges"][(a1, a2)]["min_entry_obj"] = g_1_3_min
            ret["types_view"][t]["edges"][(a1, a2)]["max_entry_obj"] = g_1_3_max
            ret["types_view"][t]["edges"][(a1, a2)]["semantics"] = "EXI=%d..%d\nENT=%d..%d" % (g_1_2_min, g_1_2_max, g_1_3_min, g_1_3_max)
            ret["types_view"][t]["edges"][(a1, a2)]["performance_events"] = values_timestamp_ee
            ret["types_view"][t]["edges"][(a1, a2)]["performance_eo"] = values_timestamp_eoe
        for edge in ret["types_view"][t]["edges"]:
            ret["types_view"][t]["edges"][edge]["events"] = len(ret["types_view"][t]["edges"][edge]["events"])
            ret["types_view"][t]["edges"][edge]["objects"] = len(ret["types_view"][t]["edges"][edge]["objects"])
            ret["types_view"][t]["edges"][edge]["eo"] = len(ret["types_view"][t]["edges"][edge]["eo"])
        for act in ret["types_view"][t]["activities"]:
            o = len(ret["types_view"][t]["activities"][act]["objects"])
            if act not in activities_mapping or activities_mapping_count[act] < o:
                activities_mapping[act] = t
                activities_mapping_count[act] = o
            ret["types_view"][t]["activities"][act]["events"] = len(ret["types_view"][t]["activities"][act]["events"])
            ret["types_view"][t]["activities"][act]["objects"] = o
            ret["types_view"][t]["activities"][act]["eo"] = len(ret["types_view"][t]["activities"][act]["eo"])
        ret["activities_mapping"] = activities_mapping

    return ret


def group_1_2(values):
    ret = {}
    for val in values:
        e1 = val[0]
        o = val[1]
        if not e1 in ret:
            ret[e1] = set()
        ret[e1].add(o)
    return list(len(x) for x in ret.values())


def group_1_3(values):
    ret = {}
    for val in values:
        e1 = val[2]
        o = val[1]
        if not e1 in ret:
            ret[e1] = set()
        ret[e1].add(o)
    return list(len(x) for x in ret.values())



from fastparquet import ParquetFile
from pm4py.objects.log.util import dataframe_utils
import deprecation
from pm4pymdl.util.parquet_importer.parameters import Parameters
from pm4py.util import exec_utils

COLUMNS = Parameters.COLUMNS.value


def apply(path, parameters=None):
    """
    Import a Parquet file

    Parameters
    -------------
    path
        Path of the file to import
    parameters
        Parameters of the algorithm, possible values:
            Parameters.COLUMNS -> columns to import from the Parquet file

    Returns
    -------------
    df
        Pandas dataframe
    """
    if parameters is None:
        parameters = {}

    columns = exec_utils.get_param_value(Parameters.COLUMNS, parameters, None)

    pf = ParquetFile(path)

    if columns:
        df = pf.to_pandas(columns)
    else:
        df = pf.to_pandas()

    return dataframe_utils.legacy_parquet_support(df)



import pandas as pd
from pm4pymdl.algo.mvp.gen_framework2 import general
from pm4pymdl.algo.mvp.utils import clean_frequency, clean_arc_frequency
from pm4py.objects.conversion.log import converter as log_conv_factory
from pm4py.objects.log.log import EventStream
from pm4py.algo.discovery.inductive import algorithm as inductive_miner
from collections import Counter
from pm4py.objects.log.util import sorting


def apply(df0, classifier_function=None, parameters=None):
    if parameters is None:
        parameters = {}

    if classifier_function is None:
        classifier_function = lambda x: x["event_activity"]

    min_acti_freq = parameters["min_acti_freq"] if "min_acti_freq" in parameters else 0
    min_edge_freq = parameters["min_edge_freq"] if "min_edge_freq" in parameters else 0

    df = df0.copy()
    df = general.preprocess(df, parameters=parameters)

    df = clean_frequency.apply(df, min_acti_freq=min_acti_freq)
    df = clean_arc_frequency.apply(df, min_freq=min_edge_freq)

    models = {}

    obj_types = [x for x in df.columns if not x.startswith("event_")]
    activities_repeated = Counter()
    activities = set()
    edges = Counter()
    start_activities = dict()
    end_activities = dict()
    acti_spec = Counter()

    for ot in obj_types:
        start_activities[ot] = set()
        end_activities[ot] = set()

        new_df = df[["event_id", "event_activity", "event_timestamp", ot]].dropna(subset=[ot])
        new_df = new_df.sort_values("event_timestamp")
        new_df = new_df.rename(
            columns={ot: "case:concept:name", "event_timestamp": "time:timestamp"})
        log = new_df.to_dict("r")
        for ev in log:
            ev["event_objtype"] = ot
            ev["concept:name"] = classifier_function(ev)
            del ev["event_objtype"]
            del ev["event_activity"]
            activities.add((ev["event_id"], ev["concept:name"]))

        log = EventStream(log)
        this_activities = set(x["concept:name"] for x in log)
        for act in this_activities:
            activities_repeated[act] += 1
        log = log_conv_factory.apply(log, variant=log_conv_factory.TO_EVENT_LOG)
        log = sorting.sort_timestamp(log, "time:timestamp")

        for trace in log:
            if trace:
                start_activities[ot].add(trace[0]["concept:name"])
                end_activities[ot].add(trace[-1]["concept:name"])
                for i in range(len(trace) - 1):
                    ev0 = trace[i]
                    ev1 = trace[i + 1]
                    edges[(ot, ev0["concept:name"], ev1["concept:name"], ev0["event_id"], ev1["event_id"], trace.attributes["concept:name"], ev0["time:timestamp"], ev1["time:timestamp"])] += 1
                    acti_spec[
                        (ot, trace[i]["concept:name"], trace[i]["event_id"], trace.attributes["concept:name"], trace[i]["time:timestamp"])] += 1
                if len(trace) > 0:
                    acti_spec[
                        (ot, trace[-1]["concept:name"], trace[-1]["event_id"], trace.attributes["concept:name"], trace[-1]["time:timestamp"])] += 1

        models[ot] = inductive_miner.apply(log, parameters=parameters)

    activities_repeated = set(x for x in activities_repeated if activities_repeated[x] > 1)
    activities = dict(Counter(list(x[1] for x in activities)))

    return {"type": "petri", "models": models, "activities": activities, "activities_repeated": activities_repeated,
            "edges": edges, "acti_spec": acti_spec}



from pm4py.visualization.common import gview
from pm4py.visualization.common import save as gsave
import tempfile
from graphviz import Digraph
import uuid

COLORS = ["#05B202", "#A13CCD", "#39F6C0", "#BA0D39", "#E90638", "#07B423", "#306A8A", "#678225", "#2742FE", "#4C9A75",
          "#4C36E9", "#7DB022", "#EDAC54", "#EAC439", "#EAC439", "#1A9C45", "#8A51C4", "#496A63", "#FB9543", "#2B49DD",
          "#13ADA5", "#2DD8C1", "#2E53D7", "#EF9B77", "#06924F", "#AC2C4D", "#82193F", "#0140D3"]

def human_readable_stat(c):
    """
    Transform a timedelta expressed in seconds into a human readable string

    Parameters
    ----------
    c
        Timedelta expressed in seconds

    Returns
    ----------
    string
        Human readable string
    """
    c = int(float(c))
    years = c // 31104000
    months = c // 2592000
    days = c // 86400
    hours = c // 3600 % 24
    minutes = c // 60 % 60
    seconds = c % 60
    if years > 0:
        return str(years) + "Y"
    if months > 0:
        return str(months) + "MO"
    if days > 0:
        return str(days) + "D"
    if hours > 0:
        return str(hours) + "h"
    if minutes > 0:
        return str(minutes) + "m"
    return str(seconds) + "s"


def apply(model, measure="frequency", freq="events", classifier="activity", projection="no", count_events=False, parameters=None):
    if parameters is None:
        parameters = {}

    if freq == "events":
        prefix = "E="
    elif freq == "objects":
        prefix = "O="
    elif freq == "eo":
        prefix = "EO="
    else:
        prefix = ""

    filename = tempfile.NamedTemporaryFile(suffix='.gv')
    viz = Digraph("pt", filename=filename.name, engine='dot', graph_attr={'bgcolor': 'transparent'})
    image_format = parameters["format"] if "format" in parameters else "png"

    min_act_freq = parameters["min_act_freq"] if "min_act_freq" in parameters else 0
    min_edge_freq = parameters["min_edge_freq"] if "min_edge_freq" in parameters else 0

    types_keys = sorted(list(model["types_view"].keys()))
    types_colors = {}

    FONTSIZE_NODES = '26'

    if freq == "semantics":
        FONTSIZE_EDGES = '19'
    else:
        FONTSIZE_EDGES = '26'


    FONTNAME_NODES = 'bold'
    FONTNAME_EDGES = 'bold'

    PENWIDTH = '4'

    for index, tk in enumerate(types_keys):
        type_color = COLORS[index % len(COLORS)]
        types_colors[tk] = type_color

    act_nodes = {}
    ordered_activities = sorted(model["activities"])
    for act in ordered_activities:
        fr_ev = model["activities"][act]["events"]
        if not freq == "semantics":
            fr = model["activities"][act][freq]
        else:
            fr = fr_ev

        if count_events:
            fr_incl = fr_ev
        else:
            fr_incl = fr

        if fr_incl >= min_act_freq:
            map_act = model["activities_mapping"][act]
            act_type_color = types_colors[map_act]
            if not freq == "semantics":
                label = act + "\n" + prefix + str(fr) + ""
            else:
                label = act
            this_uuid = str(uuid.uuid4())
            act_nodes[act] = this_uuid
            viz.node(this_uuid, label=label, color=act_type_color, shape="box", fontsize=FONTSIZE_NODES, fontname=FONTNAME_NODES)

    for index, tk in enumerate(types_keys):
        t = model["types_view"][tk]
        edges = sorted(list(t["edges"]))
        for edge in edges:
            source = edge[0]
            target = edge[1]
            source_type = model["activities_mapping"][source]
            target_type = model["activities_mapping"][target]

            if source in act_nodes and target in act_nodes:
                if projection == "no" or (projection == "source" and source_type == tk) or (projection == "target" and target_type == tk):
                    ann = t["edges"][edge][freq]
                    style = "dashed"

                    this_penwidth = PENWIDTH
                    freq_ev = t["edges"][edge]["events"]
                    if not freq == "semantics":
                        fr = t["edges"][edge][freq]
                    else:
                        fr = freq_ev

                    if count_events:
                        fr_incl = freq_ev
                    else:
                        fr_incl = fr

                    if fr_incl >= min_edge_freq:

                        if t["edges"][edge]["must_exit"]:
                            arrowtail = "box"
                        else:
                            arrowtail = None

                        if t["edges"][edge]["must_entry"]:
                            arrowhead = "normal"
                        else:
                            arrowhead = "curve"

                        if arrowtail == "box" and arrowhead == "normal":
                            style = "solid"

                        if measure == "frequency":
                            label = prefix + str(ann)
                        else:
                            if freq == "events":
                                label = "E=%s" % (human_readable_stat(t["edges"][edge]["performance_events"]))
                            elif freq == "semantics":
                                label = ""
                            else:
                                label = "EO=%s" % (human_readable_stat(t["edges"][edge]["performance_eo"]))
                        if arrowtail is not None:
                            viz.edge(act_nodes[source], act_nodes[target], style=style, label=label, color=types_colors[tk],
                                     fontcolor=types_colors[tk], fontsize=FONTSIZE_EDGES, fontname=FONTNAME_EDGES,
                                     penwidth=this_penwidth, arrowhead=arrowhead, arrowtail=arrowtail, dir="both")
                        else:
                            viz.edge(act_nodes[source], act_nodes[target], style=style, label=label, color=types_colors[tk],
                                     fontcolor=types_colors[tk], fontsize=FONTSIZE_EDGES, fontname=FONTNAME_EDGES,
                                     penwidth=this_penwidth, arrowhead=arrowhead)
    viz.format = image_format

    return viz

def save(gviz, output_file_path):
    """
    Save the diagram

    Parameters
    -----------
    gviz
        GraphViz diagram
    output_file_path
        Path where the GraphViz output should be saved
    """
    gsave.save(gviz, output_file_path)


def view(gviz):
    """
    View the diagram

    Parameters
    -----------
    gviz
        GraphViz diagram
    """
    return gview.view(gviz)



def get(dataframe, activity, attr_name, parameters=None):
    if parameters is None:
        parameters = {}

    red_df = dataframe[dataframe["event_activity"] == activity]
    return dict(red_df[attr_name].describe())



def apply(df, model, rel_ev_dict, parameters=None):
    if parameters is None:
        parameters = {}

    ret = {}
    for persp in rel_ev_dict:
        grouped_df = rel_ev_dict[persp].groupby("event_activity_merge")
        ret[persp] = list(grouped_df.size().to_dict().keys())

    return ret



import sqlite3
import os
import traceback
import pandas as pd

def apply(df, file_path, obj_df=None, parameters=None):
    if parameters is None:
        parameters = {}

    if os.path.exists(file_path):
        os.remove(file_path)

    cnx = sqlite3.connect(file_path)

    df = df.dropna(subset=["event_id"])
    df = df.dropna(subset=["event_activity"])
    df = df.dropna(subset=["event_timestamp"])
    df = df.dropna(how="all", axis=1)

    col = list(df.columns)
    rep_dict = {}
    for x in col:
        if x.startswith("event_"):
            rep_dict[x] = x.split("event_")[1]
        else:
            rep_dict[x] = "case_" + x
    df = df.rename(columns=rep_dict)

    df.to_sql(name="event_log", con=cnx, index=False)

    if obj_df is not None:
        OT = obj_df["object_type"].dropna().unique()
        dict = {"NAME": OT}
        ot_df = pd.DataFrame(dict)
        ot_df.to_sql(name="object_types", con=cnx, index=False)

        for ot in OT:
            red_df = obj_df[obj_df["object_type"] == ot].dropna(how="all", axis=1)

            red_df = red_df.dropna(subset=["object_id"]).drop(columns=["object_type"])

            col = list(red_df.columns)
            rep_dict = {}
            for x in col:
                rep_dict[x] = x.split("object_")[1]
            red_df = red_df.rename(columns=rep_dict)

            red_df.to_sql(name=ot, con=cnx, index=False)

    cnx.close()



import pyarrow as pa
import pyarrow.parquet as pq
import os
import shutil
from pm4py.objects.log.util import dataframe_utils
from enum import Enum
from pm4py.util import exec_utils


class Parameters(Enum):
    COMPRESSION = "compression"
    PARTITION_COLS = "partition_cols"
    AUTO_PARTITIONING = "auto_partitioning"
    NUM_PARTITIONS = "num_partitions"


def apply(df, path, parameters=None):
    """
    Exports a dataframe to a Parquet file

    Parameters
    ------------
    df
        Dataframe
    path
        Path
    parameters
        Possible parameters of the algorithm:
            - Parameters.COMPRESSION
            - Parameters.PARTITION_COLS
            - Parameters.AUTO_PARTITIONING
            - Parameters.NUM_PARTITIONS
    """
    if parameters is None:
        parameters = {}

    compression = exec_utils.get_param_value(Parameters.COMPRESSION, parameters, "snappy")
    partition_cols = exec_utils.get_param_value(Parameters.PARTITION_COLS, parameters, None)
    auto_partitioning = exec_utils.get_param_value(Parameters.AUTO_PARTITIONING, parameters, False)
    num_partitions = exec_utils.get_param_value(Parameters.NUM_PARTITIONS, parameters, 128)

    if partition_cols is None and auto_partitioning:
        df = dataframe_utils.insert_partitioning(df, num_partitions, parameters=parameters)
        partition_cols = ["@@partitioning"]

    df = pa.Table.from_pandas(df)

    if partition_cols is not None:
        if os.path.exists(path):
            shutil.rmtree(path)
        os.mkdir(path)
        pq.write_to_dataset(df, path, compression=compression, partition_cols=partition_cols)
    else:
        pq.write_table(df, path, compression=compression)



from pm4pymdl.algo.mvp.projection.versions import classic, activity_occurrence, group_size_hist

CLASSIC = "classic"
ACTIVITY_OCCURRENCE = "activity_occurrence"
GROUP_SIZE_HIST = "group_size_hist"

VERSIONS = {CLASSIC: classic.apply, ACTIVITY_OCCURRENCE: activity_occurrence.apply, GROUP_SIZE_HIST: group_size_hist.apply}


def apply(df, persp, variant=CLASSIC, parameters=None):
    return VERSIONS[variant](df, persp, parameters=parameters)



from os.path import dirname, join

from setuptools import setup

import pm4pymdl

def read_file(filename):
    with open(join(dirname(__file__), filename)) as f:
        return f.read()

setup(
    name=pm4pymdl.__name__,
    version=pm4pymdl.__version__,
    description=pm4pymdl.__doc__.strip(),
    long_description=read_file('README.md'),
    author=pm4pymdl.__author__,
    author_email=pm4pymdl.__author_email__,
    py_modules=[pm4pymdl.__name__],
    include_package_data=True,
    packages=['pm4pymdl', 'pm4pymdl.algo', 'pm4pymdl.algo.mvp', 'pm4pymdl.algo.mvp.utils',
              'pm4pymdl.algo.mvp.discovery', 'pm4pymdl.algo.mvp.discovery.versions', 'pm4pymdl.algo.mvp.projection',
              'pm4pymdl.algo.mvp.projection.versions', 'pm4pymdl.algo.mvp.gen_framework',
              'pm4pymdl.algo.mvp.gen_framework.util', 'pm4pymdl.algo.mvp.gen_framework.models',
              'pm4pymdl.algo.mvp.gen_framework.edge_freq', 'pm4pymdl.algo.mvp.gen_framework.node_freq',
              'pm4pymdl.algo.mvp.gen_framework.rel_events', 'pm4pymdl.algo.mvp.gen_framework.rel_activities',
              'pm4pymdl.algo.mvp.gen_framework2', 'pm4pymdl.algo.mvp.gen_framework2.md_ts',
              'pm4pymdl.algo.mvp.gen_framework2.md_dfg', 'pm4pymdl.algo.mvp.gen_framework2.md_petri_alpha',
              'pm4pymdl.algo.mvp.gen_framework2.md_process_trees',
              'pm4pymdl.algo.mvp.gen_framework2.md_petri_inductive', 'pm4pymdl.algo.mvp.gen_framework3',
              'pm4pymdl.algo.mvp.gen_framework3.versions_discovery',
              'pm4pymdl.algo.mvp.gen_framework3.versions_conformance', 'pm4pymdl.algo.mvp.gen_framework4',
              'pm4pymdl.algo.mvp.gen_framework4.versions_discovery',
              'pm4pymdl.algo.mvp.gen_framework4.versions_conformance', 'pm4pymdl.algo.mvp.get_logs_and_replay',
              'pm4pymdl.algo.mvp.get_logs_and_replay.versions', 'pm4pymdl.algo.bindings', 'pm4pymdl.algo.simulation',
              'pm4pymdl.util', 'pm4pymdl.util.parquet_exporter', 'pm4pymdl.util.parquet_exporter.versions',
              'pm4pymdl.util.parquet_importer', 'pm4pymdl.util.parquet_importer.versions', 'pm4pymdl.objects',
              'pm4pymdl.objects.jmd', 'pm4pymdl.objects.jmd.exporter', 'pm4pymdl.objects.jmd.importer',
              'pm4pymdl.objects.mdl', 'pm4pymdl.objects.mdl.exporter', 'pm4pymdl.objects.mdl.importer',
              'pm4pymdl.objects.xoc', 'pm4pymdl.objects.xoc.exporter', 'pm4pymdl.objects.xoc.exporter.versions',
              'pm4pymdl.objects.xoc.importer', 'pm4pymdl.objects.xoc.importer.versions', 'pm4pymdl.objects.sqlite',
              'pm4pymdl.objects.sqlite.exporter', 'pm4pymdl.objects.sqlite.importer', 'pm4pymdl.objects.openslex',
              'pm4pymdl.objects.openslex.importer', 'pm4pymdl.objects.openslex.importer.versions',
              'pm4pymdl.visualization', 'pm4pymdl.visualization.mvp', 'pm4pymdl.visualization.mvp.versions',
              'pm4pymdl.visualization.mvp.gen_framework', 'pm4pymdl.visualization.mvp.gen_framework.versions',
              'pm4pymdl.visualization.mvp.gen_framework2', 'pm4pymdl.visualization.mvp.gen_framework2.versions',
              'pm4pymdl.visualization.mvp.gen_framework3', 'pm4pymdl.visualization.mvp.gen_framework4',
              'pm4pymdl.visualization.petrinet', 'pm4pymdl.visualization.petrinet.versions'],
    url='http://www.pm4py.org',
    license='GPL 3.0',
    install_requires=[
        "pm4py==1.5.2.3",
        "pyarrow==1.0.1",
        "Flask",
        "flask-cors",
        "setuptools",
        "frozendict"
    ],
    project_urls={
        'Documentation': 'http://pm4py.pads.rwth-aachen.de/documentation/',
        'Source': 'https://github.com/pm4py/pm4py-source',
        'Tracker': 'https://github.com/pm4py/pm4py-source/issues',
    }
)



from copy import deepcopy

def apply(stream):
    new_stream = []

    for ev in stream:
        keys = set(ev.keys())

        event_keys = [k for k in keys if k.startswith("event_")]
        object_keys = [k for k in keys if not k in event_keys]

        basic_event = {k: ev[k] for k in event_keys}

        for k in object_keys:
            if type(ev[k]) is str:
                if ev[k][0] == "[":
                    ev[k] = eval(ev[k])
                    #ev[k] = ev[k][1:-1].split(",")
            values = ev[k]
            if values is not None:
            #if values is not None and len(values) > 0:
                if not (str(values).lower() == "nan" or str(values).lower() == "nat"):
                    for v in values:
                        event = deepcopy(basic_event)
                        event[k] = v
                        new_stream.append(event)

    return new_stream



from pm4pymdl.algo.mvp.utils import filter_metaclass
from pm4pymdl.algo.mvp.utils import succint_mdl_to_exploded_mdl


def filter_ot(df, act, ot, v1, v2, parameters=None):
    if parameters is None:
        parameters = {}

    print(len(df))

    try:
        if df.type == "succint":
            df = succint_mdl_to_exploded_mdl.apply(df)
    except:
        pass

    red_df0 = df[df["event_activity"] == act]
    red_df = red_df0.dropna(subset=[ot])

    dct = red_df.groupby("event_id").size().to_dict()
    lst = [x for x,y in dct.items() if v1 <= y <= v2]

    red_df = red_df0[red_df0["event_id"].isin(lst)]

    return filter_metaclass.do_filtering(df, red_df, parameters=parameters)




from pm4py.visualization.common import gview
from pm4py.visualization.common import save as gsave
from pm4pymdl.visualization.mvp.gen_framework2.versions import ptree, petri, dfg, ts


def apply(res, measure="frequency", freq="events", classifier="activity", projection="no", parameters=None):
    if parameters is None:
        parameters = {}

    if res["type"] == "ptree":
        return ptree.apply(res, measure=measure, freq=freq, classifier=classifier, projection=projection, parameters=parameters)
    elif res["type"] == "petri":
        return petri.apply(res, measure=measure, freq=freq, classifier=classifier, projection=projection, parameters=parameters)
    elif res["type"] == "dfg":
        return dfg.apply(res, measure=measure, freq=freq, classifier=classifier, projection=projection, parameters=parameters)
    elif res["type"] == "trans_system":
        return ts.apply(res, measure=measure, freq=freq, classifier=classifier, projection=projection, parameters=parameters)


def save(gviz, output_file_path):
    """
    Save the diagram

    Parameters
    -----------
    gviz
        GraphViz diagram
    output_file_path
        Path where the GraphViz output should be saved
    """
    gsave.save(gviz, output_file_path)


def view(gviz):
    """
    View the diagram

    Parameters
    -----------
    gviz
        GraphViz diagram
    """
    return gview.view(gviz)



from pm4pymdl.visualization.mvp.gen_framework.versions import model1 as model1_visualization
from pm4pymdl.visualization.mvp.gen_framework.versions import model2 as model2_visualization
from pm4pymdl.visualization.mvp.gen_framework.versions import model3 as model3_visualization
from pm4pymdl.visualization.mvp.gen_framework.versions import procl_der_model

from pm4pymdl.algo.mvp.gen_framework.models import model_builder as model_factory

from graphviz import Digraph
from pm4py.visualization.common import gview
from pm4py.visualization.common import save as gsave
import uuid
import tempfile

def apply(model, parameters=None):
    if parameters is None:
        parameters = {}

    if model.type == model_factory.MODEL1:
        return model1_visualization.apply(model, parameters=parameters)
    elif model.type == model_factory.MODEL2:
        return model2_visualization.apply(model, parameters=parameters)
    elif model.type == model_factory.MODEL3:
        return model3_visualization.apply(model, parameters=parameters)
    elif model.type == "proclet_derivation_model":
        return procl_der_model.apply(model, parameters=parameters)

def save(gviz, output_file_path):
    """
    Save the diagram

    Parameters
    -----------
    gviz
        GraphViz diagram
    output_file_path
        Path where the GraphViz output should be saved
    """
    gsave.save(gviz, output_file_path)


def view(gviz):
    """
    View the diagram

    Parameters
    -----------
    gviz
        GraphViz diagram
    """
    return gview.view(gviz)



import pandas as pd
from pm4pymdl.algo.mvp.utils import filter_metaclass
from pm4pymdl.algo.mvp.utils import succint_mdl_to_exploded_mdl


def apply(df, obj_type, act1, act2, minp, maxp, parameters=None):
    try:
        if df.type == "succint":
            df = succint_mdl_to_exploded_mdl.apply(df)
    except:
        pass

    if parameters is None:
        parameters = {}

    cols = [x for x in df.columns if x.startswith("event_")] + [obj_type]
    red_df = df[cols].dropna(subset=[obj_type])
    red_df = red_df.sort_values([obj_type, "event_timestamp"])
    red_df_shifted = red_df.shift(-1)
    red_df_shifted.columns = [str(col) + '_2' for col in red_df_shifted.columns]
    stacked_df = pd.concat([red_df, red_df_shifted], axis=1)
    stacked_df["@@path"] = stacked_df["event_activity"] + "," + stacked_df["event_activity_2"]
    stacked_df["@@diff"] = (stacked_df["event_timestamp_2"] - stacked_df["event_timestamp"]).astype('timedelta64[s]')
    stacked_df = stacked_df[stacked_df["@@path"] == act1+","+act2]
    #stacked_df.info()
    #print(minp, type(minp))
    #print(maxp, type(maxp))
    #print(stacked_df["@@diff"])
    stacked_df = stacked_df[minp <= stacked_df["@@diff"]]
    stacked_df = stacked_df[stacked_df["@@diff"] <= maxp]
    filt_df = red_df[red_df["event_id"].isin(stacked_df["event_id"]) | red_df["event_id"].isin(stacked_df["event_id_2"])]
    return filter_metaclass.do_filtering(df, filt_df)



from pm4pymdl.algo.mvp.gen_framework4.versions_discovery import classic

VARIANTS = {"classic": classic}


def apply(dataframe, parameters=None):
    if parameters is None:
        parameters = {}

    return VARIANTS["classic"].apply(dataframe, parameters=parameters)


def apply_stream(stream, parameters=None):
    if parameters is None:
        parameters = {}

    return VARIANTS["classic"].apply_stream(stream, parameters=parameters)


def get_stream_from_dataframe(df, parameters=None):
    if parameters is None:
        parameters = {}

    return VARIANTS["classic"].get_stream_from_dataframe(df, parameters=parameters)



class ObjCentricMultigraph(object):
    def __init__(self, dictio):
        self.dictio = dictio

    def __repr__(self):
        ret = []
        ret.append("\nObject-Centric Multigraph\n\n")
        ret.append("Activities (Events, Objects, EO):\n")
        for act in self.dictio["activities"]:
            ret.append("\t%s\tE=%d\tO=%d\tEO=%d\n" % (
                act, self.dictio["activities"][act]["events"], self.dictio["activities"][act]["objects"],
                self.dictio["activities"][act]["eo"]))
        ret.append("\nTypes:\n")
        for tk in self.dictio["types_view"]:
            ret.append("\t%s\n" % (tk))
            t = self.dictio["types_view"][tk]
            ret.append("\t\tstart activities: ")
            for act in t["start_activities"]:
                ret.append("(%s,E=%d,O=%d,EO=%d,min_obj=%d,max_obj=%d,must=%s) " % (
                    act, t["start_activities"][act]["events"], t["start_activities"][act]["objects"],
                    t["start_activities"][act]["eo"], t["start_activities"][act]["min_obj"],
                    t["start_activities"][act]["max_obj"], t["start_activities"][act]["must"]))
            ret.append("\n\t\tend activities: ")
            for act in t["end_activities"]:
                ret.append("(%s,E=%d,O=%d,EO=%d,min_obj=%d,max_obj=%d,must=%s) " % (
                    act, t["end_activities"][act]["events"], t["end_activities"][act]["objects"],
                    t["end_activities"][act]["eo"], t["end_activities"][act]["min_obj"],
                    t["end_activities"][act]["max_obj"], t["end_activities"][act]["must"]))
            ret.append("\n\t\tedges: \n")
            for edge in t["edges"]:
                ret.append("\t\t\t%s->%s E=%d O=%d EO=%d min_obj=%d max_obj=%d must=%s\n" % (
                    edge[0], edge[1], t["edges"][edge]["events"], t["edges"][edge]["objects"], t["edges"][edge]["eo"],
                    t["edges"][edge]["min_obj"], t["edges"][edge]["max_obj"], t["edges"][edge]["must"]))
        ret.append("\nMapping:\n")
        for act in self.dictio["activities_mapping"]:
            ret.append("\t%s\t%s\n" % (act, self.dictio["activities_mapping"][act]))
        return "".join(ret)



from graphviz import Digraph
import uuid
import tempfile
from pm4pymdl.visualization.mvp.gen_framework.versions import dfg_clean


COLORS = ["#05B202", "#A13CCD", "#39F6C0", "#BA0D39", "#E90638", "#07B423", "#306A8A", "#678225", "#2742FE", "#4C9A75",
          "#4C36E9", "#7DB022", "#EDAC54", "#EAC439", "#EAC439", "#1A9C45", "#8A51C4", "#496A63", "#FB9543", "#2B49DD",
          "#13ADA5", "#2DD8C1", "#2E53D7", "#EF9B77", "#06924F", "#AC2C4D", "#82193F", "#0140D3"]


def apply(model, parameters=None):
    if parameters is None:
        parameters = {}

    min_node_freq = parameters["min_node_freq"] if "min_node_freq" in parameters else 0
    min_edge_freq = parameters["min_edge_freq"] if "min_edge_freq" in parameters else 0
    dfg_cleaning_threshold = parameters["dfg_cleaning_threshold"] if "dfg_cleaning_threshold" in parameters else -1
    max_edge_ratio = parameters["max_edge_ratio"] if "max_edge_ratio" in parameters else 100000000

    image_format = "png"
    if "format" in parameters:
        image_format = parameters["format"]

    filename = tempfile.NamedTemporaryFile(suffix='.gv').name
    g = Digraph("", filename=filename, engine='dot', graph_attr={'bgcolor': 'transparent'})

    nodes_map = {}

    count_nodes = 0

    all_activities = model.get_all_activities()

    for activity in all_activities:
        this_uuid = str(uuid.uuid4())

        if model.node_freq[activity] >= min_node_freq:
            g.node(this_uuid, activity + " (" + str(model.node_freq[activity]) + ")")
            nodes_map[activity] = this_uuid

            count_nodes = count_nodes + 1

    persp_list = sorted(list(model.edge_freq))
    count_edges = 0

    for index, persp in enumerate(persp_list):
        color = COLORS[index % len(COLORS)]
        edges = model.edge_freq[persp]

        if dfg_cleaning_threshold > -1:
            edges = dfg_clean.clean_dfg_based_on_noise_thresh(edges, model.node_freq, dfg_cleaning_threshold)

        for edge in edges:
            if edge.split("@@")[0] in nodes_map and edge.split("@@")[1] in nodes_map:
                if edges[edge] >= min_edge_freq:
                    act1 = edge.split("@@")[0]
                    act2 = edge.split("@@")[1]
                    act1_corr = nodes_map[act1]
                    act2_corr = nodes_map[act2]

                    if max(edges[edge]/model.node_freq[act1], edges[edge]/model.node_freq[act2]) < max_edge_ratio:
                        g.edge(act1_corr, act2_corr, persp + " ("+str(edges[edge])+")", color=color, fontcolor=color)
                        count_edges = count_edges + 1

    if "count_nodes" in parameters and "count_edges" in parameters:
        print("Type 1 nodes",count_nodes,"recall",count_nodes/parameters["count_nodes"])
        print("Type 1 edges",count_edges,"recall",count_edges/parameters["count_edges"])
    else:
        print("Type 1 nodes",count_nodes)
        print("Type 1 edges",count_edges)

    g.attr(overlap='false')
    g.attr(fontsize='11')

    g.format = image_format

    return g



from pm4pymdl.algo.mvp.utils import succint_mdl_to_exploded_mdl
from pm4py.algo.discovery.dfg.adapters.pandas import df_statistics


def get(df):
    try:
        if df.type == "succint":
            df = succint_mdl_to_exploded_mdl.apply(df)
    except:
        pass
    activ = dict(df.groupby("event_id").first()["event_activity"].value_counts())
    max_activ_freq = max(activ.values()) if len(activ.values()) > 0 else 0
    dfg = df_statistics.get_dfg_graph(red_df, activity_key="event_activity", timestamp_key="event_timestamp",
                                      case_id_glue=persp)
    max_edge_freq = max(dfg.values()) if len(dfg) > 0 else 0
    return {"max_activ_freq": max_activ_freq, "max_edge_freq": max_edge_freq}




from pm4pymdl.algo.mvp.utils import succint_mdl_to_exploded_mdl
from pm4pymdl.algo.mvp.gen_framework3 import model
from pm4py.objects.conversion.log import converter
from collections import Counter
from statistics import median
import sys
import math


def apply(df, parameters=None):
    if parameters is None:
        parameters = {}

    stream = get_stream_from_dataframe(df, parameters=parameters)

    return apply_stream(stream, parameters=parameters)


def get_stream_from_dataframe(df, parameters=None):
    if parameters is None:
        parameters = {}

    df_type = df.type
    df = df.sort_values(["event_timestamp", "event_id"])
    if df_type == "succint":
        df = succint_mdl_to_exploded_mdl.apply(df)

    columns = [x for x in df.columns if not x.startswith("event") or x == "event_activity" or x == "event_id" or x == "event_timestamp"]
    df = df[columns]

    stream = converter.apply(df, variant=converter.Variants.TO_EVENT_STREAM)

    return stream


def apply_stream(stream, parameters=None):
    if parameters is None:
        parameters = {}

    support = parameters["support"] if "support" in parameters else 1
    epsilon = parameters["epsilon"] if "epsilon" in parameters else 0.0
    noise_obj_number = parameters["noise_obj_number"] if "noise_obj_number" in parameters else 0.0
    debug = parameters["debug"] if "debug" in parameters else False

    types_lifecycle = {}
    types_view = {}

    ret = {}

    for ev in stream:
        cl = [k for k in ev if not k.startswith("event_") and str(ev[k]) != "nan"][0]
        if not cl in types_lifecycle:
            types_lifecycle[cl] = {}
        o = ev[cl]
        if not o in types_lifecycle[cl]:
            types_lifecycle[cl][o] = []
        types_lifecycle[cl][o].append(ev)

    activities = {}

    for t in types_lifecycle:
        objects_lifecycle = types_lifecycle[t]

        activities_local = {}
        start_activities = dict()
        end_activities = dict()

        edges = {}
        for o in objects_lifecycle:
            evs = objects_lifecycle[o]
            for i in range(len(evs)):
                ev = evs[i]
                act = ev["event_activity"]
                if act not in activities:
                    activities[act] = {"events": set(), "objects": set(), "eo": set()}
                if act not in activities_local:
                    activities_local[act] = {"objects": set(), "eo": set(), "preceded_by": dict(),
                                             "followed_by": dict()}
                activities[act]["events"].add(ev["event_id"])
                activities[act]["objects"].add(ev[t])
                activities[act]["eo"].add((ev["event_id"], ev[t]))
                activities_local[act]["objects"].add(ev[t])
                activities_local[act]["eo"].add((ev["event_id"], ev[t]))
                if i == 0:
                    if act not in start_activities:
                        start_activities[act] = {"events": set(), "objects": set(), "eo": set(), "must": False,
                                                 "min_obj": 1, "max_obj": sys.maxsize}
                    start_activities[act]["events"].add(ev["event_id"])
                    start_activities[act]["objects"].add(ev[t])
                    start_activities[act]["eo"].add((ev["event_id"], ev[t]))
                if i > 0:
                    if not evs[i - 1]["event_activity"] in activities_local[act]["preceded_by"]:
                        activities_local[act]["preceded_by"][evs[i - 1]["event_activity"]] = set()
                    activities_local[act]["preceded_by"][evs[i - 1]["event_activity"]].add((ev["event_id"], ev[t]))
                if i < len(evs) - 1:
                    if not evs[i + 1]["event_activity"] in activities_local[act]["followed_by"]:
                        activities_local[act]["followed_by"][evs[i + 1]["event_activity"]] = set()
                    activities_local[act]["followed_by"][evs[i + 1]["event_activity"]].add((ev["event_id"], ev[t]))
                if i == len(evs) - 1:
                    if act not in end_activities:
                        end_activities[act] = {"events": set(), "objects": set(), "eo": set(), "must": False,
                                               "min_obj": 1, "max_obj": sys.maxsize}
                    end_activities[act]["events"].add(ev["event_id"])
                    end_activities[act]["objects"].add(ev[t])
                    end_activities[act]["eo"].add((ev["event_id"], ev[t]))
            for i in range(len(evs) - 1):
                ev0 = evs[i]
                ev1 = evs[i + 1]
                if (ev0["event_activity"], ev1["event_activity"]) not in edges:
                    edges[((ev0["event_activity"], ev1["event_activity"]))] = {"events": set(), "objects": set(),
                                                                               "eo": set(), "eeo": set(),
                                                                               "min_obj": 1, "max_obj": sys.maxsize,
                                                                               "must": False, "performance_events": set(),
                                                                               "performance_eo": []}
                edges[(ev0["event_activity"], ev1["event_activity"])]["events"].add(ev1["event_id"])
                edges[(ev0["event_activity"], ev1["event_activity"])]["objects"].add(ev1[t])
                edges[(ev0["event_activity"], ev1["event_activity"])]["eo"].add((ev1["event_id"], ev1[t]))
                edges[(ev0["event_activity"], ev1["event_activity"])]["eeo"].add(
                    (str(ev0["event_id"]) + "@@@" + str(ev1["event_id"]), ev1[t]))
                diff = ev1["event_timestamp"].timestamp() - ev0["event_timestamp"].timestamp()
                edges[(ev0["event_activity"], ev1["event_activity"])]["performance_events"].add((ev1["event_id"], diff))
                edges[(ev0["event_activity"], ev1["event_activity"])]["performance_eo"].append(diff)

        for edge in edges:
            eo_dict = {}
            for eo in edges[edge]["eeo"]:
                eo_dict[eo[0]] = set()
            for eo in edges[edge]["eeo"]:
                eo_dict[eo[0]].add(eo[1])
            for e in eo_dict:
                eo_dict[e] = len(eo_dict[e])

            all_values = sorted(list(eo_dict.values()))
            if all_values:
                all_values = all_values[:math.ceil(len(all_values)*(1.0-noise_obj_number))]
            min_obj = min(all_values)
            max_obj = max(all_values)

            if not edge[0] == edge[1]:
                predecessors = sorted(list(activities_local[edge[1]]["preceded_by"]), reverse=True,
                                      key=lambda x: len(activities_local[edge[1]]["preceded_by"][x]))
                num = activities_local[edge[1]]["preceded_by"][predecessors[0]]
                den = activities_local[edge[1]]["preceded_by"][predecessors[0]]

                for i in range(1, len(predecessors)):
                    den = den.union(activities_local[edge[1]]["preceded_by"][predecessors[i]])

                q0 = 1 - len(num) / len(den) if len(den) > 0 else sys.maxsize

                if q0 <= epsilon:
                    set_difference = edges[edge]["eo"].difference(
                        activities_local[edge[1]]["preceded_by"][predecessors[0]])
                    set_intersection = edges[edge]["eo"].intersection(
                        activities_local[edge[1]]["preceded_by"][predecessors[0]])

                    q1 = len(set_intersection)
                    q2 = len(set_difference) / len(set_intersection) if len(set_intersection) > 0 else sys.maxsize

                    if debug:
                        print("edge ", edge, "q1=", q1, "q2=", q2)
                    if q1 >= support and q2 <= epsilon:
                        edges[edge]["must"] = True

            edges[edge]["min_obj"] = min_obj
            edges[edge]["max_obj"] = max_obj
            edges[edge]["events"] = len(edges[edge]["events"])
            edges[edge]["objects"] = len(edges[edge]["objects"])
            edges[edge]["eo"] = len(edges[edge]["eo"])
            edges[edge]["eeo"] = len(edges[edge]["eeo"])

            edges[edge]["semantics"] = "%d..%d" % (edges[edge]["min_obj"], edges[edge]["max_obj"])
            if edges[edge]["must"]:
                edges[edge]["semantics"] = "(M) " + edges[edge]["semantics"]

            edges[edge]["performance_eo"] = median(edges[edge]["performance_eo"])
            edges[edge]["performance_events"] = list(x[1] for x in edges[edge]["performance_events"])
            edges[edge]["performance_events"] = median(edges[edge]["performance_events"])

        max_start_act_occ = max(len(start_activities[x]["events"]) for x in start_activities)
        start_activities_within_support = {x: y for x, y in start_activities.items() if len(y["events"]) >= max_start_act_occ * epsilon}

        for act in start_activities:
            eo_dict = {}
            for eo in start_activities[act]["eo"]:
                eo_dict[eo[0]] = set()
            for eo in start_activities[act]["eo"]:
                eo_dict[eo[0]].add(eo[1])
            for e in eo_dict:
                eo_dict[e] = len(eo_dict[e])


            all_values = sorted(list(eo_dict.values()))
            if all_values:
                all_values = all_values[:math.ceil(len(all_values)*(1.0-noise_obj_number))]
            min_obj = min(all_values)
            max_obj = max(all_values)

            start_activities[act]["min_obj"] = min_obj
            start_activities[act]["max_obj"] = max_obj

            set_intersection = start_activities[act]["eo"].intersection(activities_local[act]["eo"])
            set_difference = start_activities[act]["eo"].difference(activities_local[act]["eo"])

            q1 = len(set_intersection)
            q2 = len(set_difference) / len(set_intersection) if len(set_intersection) > 0 else sys.maxsize

            if debug:
                print("start_activity ", act, "q1=", q1, "q2=", q2)
            if q1 >= support and q2 <= epsilon and len(start_activities_within_support) == 1:
                start_activities[act]["must"] = True

            start_activities[act]["events"] = len(start_activities[act]["events"])
            start_activities[act]["objects"] = len(start_activities[act]["objects"])
            start_activities[act]["eo"] = len(start_activities[act]["eo"])
            start_activities[act]["semantics"] = "%d..%d" % (start_activities[act]["min_obj"], start_activities[act]["max_obj"])
            if start_activities[act]["must"]:
                start_activities[act]["semantics"] = "(M) " + start_activities[act]["semantics"]

        max_end_act_occ = max(len(end_activities[x]["events"]) for x in end_activities)
        end_activities_within_support = {x: y for x, y in end_activities.items()  if len(y["events"]) >= max_end_act_occ * epsilon}

        for act in end_activities:
            eo_dict = {}
            for eo in end_activities[act]["eo"]:
                eo_dict[eo[0]] = set()
            for eo in end_activities[act]["eo"]:
                eo_dict[eo[0]].add(eo[1])
            for e in eo_dict:
                eo_dict[e] = len(eo_dict[e])

            all_values = sorted(list(eo_dict.values()))
            if all_values:
                all_values = all_values[:math.ceil(len(all_values)*(1.0-noise_obj_number))]
            min_obj = min(all_values)
            max_obj = max(all_values)

            end_activities[act]["min_obj"] = min_obj
            end_activities[act]["max_obj"] = max_obj

            set_intersection = activities_local[act]["eo"].intersection(end_activities[act]["eo"])
            set_difference = activities_local[act]["eo"].difference(end_activities[act]["eo"])

            q1 = len(set_intersection)
            q2 = len(set_difference) / len(set_intersection) if len(set_intersection) > 0 else sys.maxsize

            if debug:
                print("end_activity ", act, "q1=", q1, "q2=", q2)
            if q1 >= support and q2 <= epsilon and len(end_activities_within_support) == 1:
                end_activities[act]["must"] = True

            end_activities[act]["events"] = len(end_activities[act]["events"])
            end_activities[act]["objects"] = len(end_activities[act]["objects"])
            end_activities[act]["eo"] = len(end_activities[act]["eo"])
            end_activities[act]["semantics"] = "%d..%d" % (end_activities[act]["min_obj"], end_activities[act]["max_obj"])
            if end_activities[act]["must"]:
                end_activities[act]["semantics"] = "(M) " + end_activities[act]["semantics"]

        types_view[t] = {"start_activities": start_activities, "end_activities": end_activities,
                         "activities_local": activities_local, "edges": edges}

    for act in activities:
        activities[act]["events"] = len(activities[act]["events"])
        activities[act]["objects"] = len(activities[act]["objects"])
        activities[act]["eo"] = len(activities[act]["eo"])

    activities_mapping = {}
    for t in types_view:
        for act in types_view[t]["activities_local"]:
            act_detail = types_view[t]["activities_local"][act]
            if not act in activities_mapping or activities_mapping[act][1] < len(act_detail["eo"]):
                activities_mapping[act] = (t, len(act_detail["eo"]))
    ret["activities"] = activities
    ret["types_view"] = types_view
    ret["activities_mapping"] = activities_mapping

    return model.ObjCentricMultigraph(ret)



from collections import Counter
from statistics import mean


def projection(edges_map, reference_map, type="no", return_acti_map=False):
    acti_assignation_map = {}
    acti_assignation_score = {}
    for k in reference_map:
        for a in reference_map[k]:
            if not a in acti_assignation_map or reference_map[k][a] > acti_assignation_score[a]:
                acti_assignation_map[a] = k
                acti_assignation_score[a] = reference_map[k][a]

    if not type == "no":
        for k in edges_map:
            if type == "source":
                edges_map[k] = {x: y for x, y in edges_map[k].items() if acti_assignation_map[x[0]] == k}
            elif type == "target":
                edges_map[k] = {x: y for x, y in edges_map[k].items() if acti_assignation_map[x[1]] == k}

    if return_acti_map:
        return edges_map, acti_assignation_map
    return edges_map


def projection_edges_freq(edges_map, events_map, min_edge_freq):
    for k in events_map:
        for k2 in events_map[k]:
            if events_map[k][k2] < min_edge_freq:
                del edges_map[k][k2]

    return edges_map


def human_readable_stat(c):
    """
    Transform a timedelta expressed in seconds into a human readable string

    Parameters
    ----------
    c
        Timedelta expressed in seconds

    Returns
    ----------
    string
        Human readable string
    """
    c = int(float(c))
    years = c // 31104000
    months = c // 2592000
    days = c // 86400
    hours = c // 3600 % 24
    minutes = c // 60 % 60
    seconds = c % 60
    if years > 0:
        return str(years) + "Y"
    if months > 0:
        return str(months) + "MO"
    if days > 0:
        return str(days) + "D"
    if hours > 0:
        return str(hours) + "h"
    if minutes > 0:
        return str(minutes) + "m"
    return str(seconds) + "s"


def get_events_edges_map_performance(key, res):
    edges = [x for x in res["edges"] if x[0] == key]
    edges_map = {}
    for x in edges:
        k = (x[1], x[2])
        if k not in edges_map:
            edges_map[k] = dict()
        edges_map[k][(x[3], x[4])] = x[7].timestamp() - x[6].timestamp()
    for k in edges_map:
        edges_map[k] = mean(edges_map[k].values())
    return edges_map


def get_objects_edges_map_performance(key, res):
    edges = [x for x in res["edges"] if x[0] == key]
    edges_map = {}
    for x in edges:
        k = (x[1], x[2])
        if k not in edges_map:
            edges_map[k] = dict()
        edges_map[k][x[5]] = x[7].timestamp() - x[6].timestamp()
    for k in edges_map:
        edges_map[k] = mean(edges_map[k].values())
    return edges_map


def get_eo_edges_map_performance(key, res):
    edges = [x for x in res["edges"] if x[0] == key]
    edges_map = {}
    for x in edges:
        k = (x[1], x[2])
        if k not in edges_map:
            edges_map[k] = list()
        edges_map[k].append(x[7].timestamp() - x[6].timestamp())
    for k in edges_map:
        edges_map[k] = mean(edges_map[k])
    return edges_map


def get_edges_map_performance(key, res, variant="events"):
    if variant == "events":
        return get_events_edges_map_performance(key, res)
    elif variant == "eo":
        return get_eo_edges_map_performance(key, res)
    elif variant == "objects":
        return get_objects_edges_map_performance(key, res)


def get_events_edges_map_frequency(key, res):
    edges = [x for x in res["edges"] if x[0] == key]
    edges_map = {}
    for x in edges:
        k = (x[1], x[2])
        if k not in edges_map:
            edges_map[k] = Counter()
        edges_map[k][(x[3], x[4])] += res["edges"][x]
    for k in edges_map:
        edges_map[k] = len(edges_map[k])
    return edges_map


def get_objects_edges_map_frequency(key, res):
    edges = [x for x in res["edges"] if x[0] == key]
    edges_map = {}
    for x in edges:
        k = (x[1], x[2])
        if k not in edges_map:
            edges_map[k] = Counter()
        edges_map[k][x[5]] += res["edges"][x]
    for k in edges_map:
        edges_map[k] = len(edges_map[k])
    return edges_map


def get_eo_edges_map_frequency(key, res):
    edges = [x for x in res["edges"] if x[0] == key]
    edges_map = {}
    for x in edges:
        k = (x[1], x[2])
        if k not in edges_map:
            edges_map[k] = Counter()
        edges_map[k][(x[3], x[4])] += res["edges"][x]
    for k in edges_map:
        edges_map[k] = sum(edges_map[k].values())
    return edges_map


def get_edges_map_frequency(key, res, variant="events"):
    if variant == "events":
        return get_events_edges_map_frequency(key, res)
    elif variant == "eo":
        return get_eo_edges_map_frequency(key, res)
    elif variant == "objects":
        return get_objects_edges_map_frequency(key, res)


def get_edges_map(key, res, measure="frequency", variant="events"):
    if measure == "frequency":
        return get_edges_map_frequency(key, res, variant=variant)
    else:
        return get_edges_map_performance(key, res, variant=variant)


def get_activity_map_events_frequency(key, res):
    activities = [x for x in res["acti_spec"] if x[0] == key]
    activities_map = {}
    for x in activities:
        k = x[1]
        if k not in activities_map:
            activities_map[k] = Counter()
        activities_map[k][x[2]] += res["acti_spec"][x]
    for k in activities_map:
        activities_map[k] = len(activities_map[k])
    return activities_map


def get_activity_map_eo_frequency(key, res):
    activities = [x for x in res["acti_spec"] if x[0] == key]
    activities_map = {}
    for x in activities:
        k = x[1]
        if k not in activities_map:
            activities_map[k] = Counter()
        activities_map[k][x[2]] += res["acti_spec"][x]
    for k in activities_map:
        activities_map[k] = sum(activities_map[k].values())
    return activities_map


def get_activity_map_objects_frequency(key, res):
    activities = [x for x in res["acti_spec"] if x[0] == key]
    activities_map = {}
    for x in activities:
        k = x[1]
        if k not in activities_map:
            activities_map[k] = Counter()
        activities_map[k][x[3]] += res["acti_spec"][x]
    for k in activities_map:
        activities_map[k] = len(activities_map[k])
    return activities_map


def get_activity_map_frequency(key, res, variant="events"):
    if variant == "events":
        return get_activity_map_events_frequency(key, res)
    elif variant == "eo":
        return get_activity_map_eo_frequency(key, res)
    elif variant == "objects":
        return get_activity_map_objects_frequency(key, res)



from pm4pymdl.algo.mvp.utils import succint_mdl_to_exploded_mdl
from pm4py.objects.conversion.log import converter as log_conv_factory


def apply(df, persp, parameters=None):
    if parameters is None:
        parameters = {}

    try:
        if df.type == "succint":
            df = succint_mdl_to_exploded_mdl(df)
            df.type = "exploded"
    except:
        pass

    cols = [x for x in df.columns if x.startswith("event_") or x == persp]
    df = df[cols].dropna(subset=[persp])

    renaming = {}
    renaming["event_activity"] = "concept:name"
    renaming["event_timestamp"] = "time:timestamp"
    renaming[persp] = "case:concept:name"

    df = df.rename(columns=renaming)
    df = df.dropna(subset=["concept:name"])

    log = log_conv_factory.apply(df)

    return log




import os
import pandas as pd

from pm4py.util import constants, exec_utils
from pm4py.util import xes_constants as xes
from enum import Enum

import deprecation
import traceback

VERSIONS = dict()
VERSIONS_LOG = dict()

DEFAULT_VARIANT = None
DEFAULT_VARIANT_LOG = None

Variants = None

try:
    from pm4pymdl.util.parquet_importer.versions import fastparquet

    VERSIONS["fastparquet"] = fastparquet
except:
    # Fastparquet is not installed
    # traceback.print_exc()
    pass

try:
    from pm4pymdl.util.parquet_importer.versions import pyarrow

    VERSIONS["pyarrow"] = pyarrow
    VERSIONS_LOG["pyarrow"] = pyarrow
except:
    # Pyarrow is not installed
    # traceback.print_exc()
    pass


class Parameters(Enum):
    COLUMNS = "columns"


# define the Variants enum dynamically (since when defined cannot be changed)
versions_keys = sorted(list(VERSIONS.keys()))
if versions_keys == ["fastparquet", "pyarrow"]:
    class Variants(Enum):
        FASTPARQUET = VERSIONS["fastparquet"]
        PYARROW = VERSIONS["pyarrow"]


    DEFAULT_VARIANT = Variants.PYARROW
    DEFAULT_VARIANT_LOG = Variants.PYARROW
elif versions_keys == ["pyarrow"]:
    class Variants(Enum):
        PYARROW = VERSIONS["pyarrow"]


    DEFAULT_VARIANT = Variants.PYARROW
    DEFAULT_VARIANT_LOG = Variants.PYARROW
elif versions_keys == ["fastparquet"]:
    class Variants(Enum):
        FASTPARQUET = VERSIONS["fastparquet"]


    DEFAULT_VARIANT = Variants.FASTPARQUET

COLUMNS = Parameters.COLUMNS.value


def apply(path, parameters=None, variant=DEFAULT_VARIANT):
    """
    Import a Parquet file

    Parameters
    -------------
    path
        Path of the file to import
    parameters
        Parameters of the algorithm, possible values:
            Parameters.COLUMNS -> columns to import from the Parquet file
    variant
        Variant of the algorithm, possible values:
            - Variants.PYARROW
            - Variants.FASTPARQUET

    Returns
    -------------
    df
        Pandas dataframe
    """
    if parameters is None:
        parameters = {}

    timestamp_key = parameters[
        constants.PARAMETER_CONSTANT_TIMESTAMP_KEY] if constants.PARAMETER_CONSTANT_TIMESTAMP_KEY in parameters else xes.DEFAULT_TIMESTAMP_KEY
    case_id_key = parameters[
        constants.PARAMETER_CONSTANT_CASEID_KEY] if constants.PARAMETER_CONSTANT_CASEID_KEY in parameters else constants.CASE_CONCEPT_NAME

    if os.path.isdir(path):
        all_parquets = get_list_parquet(path, parameters=parameters)
        dataframes = []
        for file in all_parquets:
            dataframes.append(exec_utils.get_variant(variant).apply(file, parameters=parameters))
        df = pd.concat(dataframes)
        df = df.sort_values([case_id_key, timestamp_key])
    else:
        df = exec_utils.get_variant(variant).apply(path, parameters=parameters)

    return df


def import_log(path, parameters=None, variant=DEFAULT_VARIANT_LOG):
    """
    Import a Parquet file

    Parameters
    -------------
    path
        Path of the file to import
    parameters
        Parameters of the algorithm, possible values:
            Parameters.COLUMNS -> columns to import from the Parquet file
    variant
        Variant of the algorithm, possible values:
            - Variants.PYARROW

    Returns
    -------------
    df
        Pandas dataframe
    """
    if parameters is None:
        parameters = {}

    return exec_utils.get_variant(variant).import_log(path, parameters=parameters)


def import_minimal_log(path, parameters=None, variant=DEFAULT_VARIANT_LOG):
    """
    Import a Parquet file (as a minimal log with only the essential columns)

    Parameters
    -------------
    path
        Path of the file to import
    parameters
        Parameters of the algorithm, possible values:
            Parameters.COLUMNS -> columns to import from the Parquet file
    variant
        Variant of the algorithm, possible values:
            - Variants.PYARROW

    Returns
    -------------
    df
        Pandas dataframe
    """
    if parameters is None:
        parameters = {}

    parameters[COLUMNS] = [constants.CASE_CONCEPT_NAME, xes.DEFAULT_NAME_KEY, xes.DEFAULT_TIMESTAMP_KEY]

    return exec_utils.get_variant(variant).import_log(path, parameters=parameters)


def import_df(path, parameters=None, variant=DEFAULT_VARIANT):
    """
    Import a Parquet file

    Parameters
    -------------
    path
        Path of the file to import
    parameters
        Parameters of the algorithm, possible values:
            Parameters.COLUMNS -> columns to import from the Parquet file
    variant
        Variant of the algorithm, possible values:
            - Variants.PYARROW

    Returns
    -------------
    df
        Pandas dataframe
    """
    return apply(path, variant=variant, parameters=parameters)


def get_list_parquet(path, parameters=None):
    """
    Gets the list of Parquets contained into a dataset

    Parameters
    -------------
    path
        Path where the dataset is
    parameters
        Possible parameters of the algorithm

    Returns
    -------------
    paths
        List of paths
    """
    if parameters is None:
        parameters = {}

    paths = []
    subelements = os.listdir(path)
    for subel in subelements:
        if os.path.isdir(os.path.join(path, subel)):
            sub2 = os.listdir(os.path.join(path, subel))
            for subel2 in sub2:
                if subel2.endswith(".parquet"):
                    if os.path.isfile(os.path.join(path, subel, subel2)):
                        paths.append(os.path.join(path, subel, subel2))
        else:
            if subel.endswith(".parquet"):
                if os.path.isfile(os.path.join(path, subel)):
                    paths.append(os.path.join(path, subel))
    return paths



import tempfile
import uuid
from graphviz import Digraph
from pm4pymdl.visualization.mvp.gen_framework2.versions import util

COLORS = ["#05B202", "#A13CCD", "#39F6C0", "#BA0D39", "#E90638", "#07B423", "#306A8A", "#678225", "#2742FE", "#4C9A75",
          "#4C36E9", "#7DB022", "#EDAC54", "#EAC439", "#EAC439", "#1A9C45", "#8A51C4", "#496A63", "#FB9543", "#2B49DD",
          "#13ADA5", "#2DD8C1", "#2E53D7", "#EF9B77", "#06924F", "#AC2C4D", "#82193F", "#0140D3"]


def get_label(act):
    return act

def apply(res, measure="frequency", freq="events", classifier="activity", projection="no", parameters=None):
    if parameters is None:
        parameters = {}

    min_edge_freq = parameters["min_edge_freq"] if "min_edge_freq" in parameters else 0

    filename = tempfile.NamedTemporaryFile(suffix='.gv')
    viz = Digraph("pt", filename=filename.name, engine='dot', graph_attr={'bgcolor': 'transparent'})
    image_format = parameters["format"] if "format" in parameters else "png"
    acti_map = {}

    freq_prefix = "E="
    if freq == "objects":
        freq_prefix = "O="
    elif freq == "eo":
        freq_prefix = "EO="

    node_shape = "box" if classifier == "activity" else "box3d"

    reference_map = {}
    events_map = {}

    edges_map = {}
    activ_freq_map = {}

    for key in res["models"]:
        edges_map[key] = util.get_edges_map(key, res, measure=measure, variant=freq)
        activ_freq_map[key] = util.get_activity_map_frequency(key, res, variant=freq)
        reference_map[key] = util.get_activity_map_frequency(key, res, variant="objects")
        events_map[key] = util.get_edges_map(key, res, measure="frequency", variant="events")

    edges_map = util.projection_edges_freq(edges_map, events_map, min_edge_freq)
    for k in edges_map:
        if len(edges_map[k]) == 0:
            if k in reference_map:
                del reference_map[k]
    edges_map, acti_assignation_map = util.projection(edges_map, reference_map, type=projection, return_acti_map=True)

    count = 0
    for key, model in res["models"].items():
        if len(events_map[key]) > 0:
            persp_color = COLORS[count % len(COLORS)]
            count = count + 1

            for act in model:
                act_id = str(id(act))

                if key == acti_assignation_map[act]:
                    acti_map[act] = act_id
                    if act in res["activities_repeated"]:
                        label = get_label(act)
                        label = get_label(act)+"\n("+freq_prefix+str(activ_freq_map[key][act])+")"
                        viz.node(act_id, label, style='filled', fillcolor="white", color=persp_color, shape=node_shape, width='1.3')
                    else:
                        label = get_label(act)
                        label = get_label(act)+"\n("+freq_prefix+str(activ_freq_map[key][act])+")"
                        viz.node(act_id, label, style='filled', fillcolor=persp_color, shape=node_shape, width='1.3')
    count = 0
    for key, model in res["models"].items():
        if len(events_map[key]) > 0:
            persp_color = COLORS[count % len(COLORS)]
            count = count + 1

            sn_uuid = str(uuid.uuid4())
            viz.node(sn_uuid, str(key), color=persp_color, fontcolor=persp_color, shape='underline', fixedsize='true', width='0.75')
            en_uuid = str(uuid.uuid4())
            viz.node(en_uuid, str(key), color=persp_color, fontcolor=persp_color, shape='none', fixedsize='true', width='0.75')

            for act in model:
                if act in res["start_activities"][key]:
                    viz.edge(sn_uuid, acti_map[act], color=persp_color)

                if act in res["end_activities"][key]:
                    viz.edge(acti_map[act], en_uuid, color=persp_color)

    count = 0
    for key, model in res["models"].items():
        if len(events_map[key]) > 0:
            persp_color = COLORS[count % len(COLORS)]
            count = count + 1

            for k in edges_map[key]:
                if measure == "frequency":
                    viz.edge(acti_map[k[0]], acti_map[k[1]], xlabel=freq_prefix+str(edges_map[key][k]), color=persp_color, fontcolor=persp_color)
                elif measure == "performance":
                    viz.edge(acti_map[k[0]], acti_map[k[1]], xlabel=freq_prefix+"P="+util.human_readable_stat(edges_map[key][k]), color=persp_color, fontcolor=persp_color)

    viz.attr(overlap='false')
    viz.attr(fontsize='11')
    viz.format = image_format

    return viz
